###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
""" Help Command for RDMC """

try:
    from rdmc_base_classes import RdmcCommandBase, RdmcOptionParser
except ImportError:
    from ilorest.rdmc_base_classes import RdmcCommandBase, RdmcOptionParser
try:
    from rdmc_helper import (
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        ReturnCodes,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        ReturnCodes,
    )


class HelpCommand(RdmcCommandBase):
    """Constructor"""

    def __init__(self):
        self.ident = {
            "name": "help",
            "usage": None,
            "description": "help [COMMAND]\n\tFor more detailed command descriptions"
            " use the help command feature\n\texample: help login\n",
            "summary": "Displays command line syntax and help menus for individual commands." " Example: help login\n",
            "aliases": [],
            "auxcommands": [],
        }
        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()

    def run(self, line, help_disp=False):
        """Wrapper function for help main function
        :param line: command line input
        :type line: string.
        """
        if help_disp:
            self.parser.print_help()
            return ReturnCodes.SUCCESS
        try:
            (options, args) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")

        if not args or not line:
            RdmcOptionParser().print_help()
            if self.rdmc:
                cmddict = self.rdmc.get_commands()
                sorted_keys = sorted(list(cmddict.keys()))

                for key in sorted_keys:
                    if key[0] == "_":
                        continue
                    else:
                        self.rdmc.ui.printer("\n%s\n" % key)
                    for cmd in cmddict[key]:
                        self.rdmc.ui.printer(
                            "%-25s - %s\n"
                            % (
                                self.rdmc.commands_dict[cmd].ident["name"],
                                self.rdmc.commands_dict[cmd].ident["summary"],
                            )
                        )
        else:
            if self.rdmc:
                cmddict = self.rdmc.get_commands()
                sorted_keys = list(cmddict.keys())

                for key in sorted_keys:
                    for cmd in cmddict[key]:
                        cmd_s = cmd.split("Command")
                        cmd_s = cmd_s[0]
                        if args[0].lower() == cmd_s.lower():
                            self.rdmc.ui.printer(self.rdmc.commands_dict[cmd].ident["description"] + "\n")
                            return ReturnCodes.SUCCESS
                raise InvalidCommandLineError("Command '%s' not found." % args[0])
        # Return code
        return ReturnCodes.SUCCESS

    def definearguments(self, customparser):
        """Wrapper function for new command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return



###
# Copyright 2017 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""Base implementation for cli interaction with Redfish interface"""

# ---------Imports---------

import getpass
import os
import re
import subprocess
import sys

try:
    from rdmc_helper import UI
except ImportError:
    from ilorest.rdmc_helper import UI

if os.name == "nt":
    import ctypes
    from ctypes import windll, wintypes

# ---------End of imports---------


class CommandNotFoundException(Exception):
    """Exception throw when Command is not found"""

    pass


class ResourceAllocationError(Exception):
    """Exception throw when Command is not found"""

    pass


def get_user_config_dir():
    """Platform specific directory for user configuration.

    :returns: returns the user configuration directory.
    :rtype: string
    """
    if os.name == "nt":
        try:
            csidl_appdata = 26

            shgetfolderpath = windll.shell32.SHGetFolderPathW
            shgetfolderpath.argtypes = [
                wintypes.HWND,
                ctypes.c_int,
                wintypes.HANDLE,
                wintypes.DWORD,
                wintypes.LPCWSTR,
            ]

            path_buf = ctypes.create_unicode_buffer(wintypes.MAX_PATH)
            result = shgetfolderpath(0, csidl_appdata, 0, 0, path_buf)

            if result == 0:
                return path_buf.value
        except ImportError:
            pass

        return os.environ["APPDATA"]

    return os.path.expanduser("~")


def is_exe(filename):
    """Determine if filename is an executable.

    :param filename: the filename to examine.
    :type filename: string
    :returns: True if filename is executable False otherwise.
    :rtype: boolean
    """
    if sys.platform == "win32":
        if os.path.exists(filename):
            return True

        # windows uses PATHEXT to list valid executable extensions
        pathext = os.environ["PATHEXT"].split(os.pathsep)
        (_, ext) = os.path.splitext(filename)

        if ext in pathext:
            return True
    else:
        if os.path.exists(filename) and os.access(filename, os.X_OK):
            return True

    return False


def find_exe(filename, path=None):
    """Search path for a executable (aka which)

    :param filename: the filename to search for.
    :type filename: string.
    :param path: the path(s) to search (default: os.environ['PATH'])
    :param path: string separated with os.pathsep
    :returns: string with full path to the file or None if not found.
    :rtype: string or None.
    """
    if path is None:
        path = os.environ["PATH"]

    pathlist = path.split(os.pathsep)

    # handle fully qualified paths
    if os.path.isfile(filename) and is_exe(filename):
        return filename

    for innerpath in pathlist:
        foundpath = os.path.join(innerpath, filename)

        if is_exe(foundpath):
            return foundpath

    return None


def get_terminal_size():
    """Returns the rows and columns of the terminal as a tuple.

    :returns: the row and column count of the terminal
    :rtype: tuple (cols, rows)
    """
    _tuple = (80, 25)  # default

    if os.name == "nt":
        pass
    else:
        which_stty = find_exe("stty")

        if which_stty:
            args = [which_stty, "size"]
            try:
                procs = subprocess.Popen(args, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            except OSError as excp:
                raise ResourceAllocationError(str(excp))
            (stdout_s, _) = procs.communicate()

            _ = procs.wait()
            # python3 change
            if isinstance(stdout_s, bytes):
                stdout_s = stdout_s.decode("utf-8")
            if stdout_s and re.search(r"^\d+ \d+$", stdout_s):
                rows, cols = stdout_s.split()
                _tuple = (cols, rows)

    return _tuple


class CLI(object):
    """Class for building command line interfaces."""

    def __init__(self, verbosity=1, out=sys.stdout):
        self._verbosity = verbosity
        self._out = out
        cols, rows = get_terminal_size()
        self._cols = int(cols)
        self._rows = int(rows)

    def verbosity(self, verbosity):
        self._verbosity = verbosity

    def get_hrstr(self, character="-"):
        """returns a string suitable for use as a horizontal rule.

        :param character: the character to use as the rule. (default -)
        :type character: str.
        """
        return "%s\n" % (character * self._cols)

    def printer(self, data, flush=True):
        """printing wrapper

        :param data: data to be printed to the output stream
        :type data: str.
        :param flush: flush buffer - True, not flush output buffer - False
        :type flush: boolean
        """

        UI(self._verbosity).printer(data, flush)

    def horizontalrule(self, character="-"):
        """writes a horizontal rule to the file handle.

        :param character: the character to use as the rule. (default -)
        :type character: str.
        """

        self.printer(self.get_hrstr(character=character))

    def version(self, progname, version, extracontent):
        """Prints a version string to fileh.

        :param progname: the name of the program.
        :type progname: str.
        :param version: the version of the program.
        :type version str.
        :param fileh: the file handle to write to. Default is sys.stdout.
        :type fileh: file object
        :returns: None
        """
        tmp = "%s version %s\n%s" % (progname, version, extracontent)
        self.printer(tmp)
        self.horizontalrule()

    def prompt_password(self, msg, default=None):
        """Convenient password prompting function

        :param msg: prompt text.
        :type msg: str.
        :param default: default value if user does not enter anything.
        :type default: str.
        :returns: string user entered, or default if nothing entered
        """
        message = "%s : " % msg

        if default is not None:
            message = "%s [%s] : " % (msg, default)

        i = getpass.getpass(message)

        if i is None or len(i) == 0:
            i = default

        i = str(i)

        return i.strip()



"""
This module contains tests for PmemHelpers
"""
from __future__ import absolute_import

import pytest

from ..conftest import PmemHelpers


class TestPmemHelpers(object):
    """
    Class to test functions in PmemHelpers
    """
    mem_members = [
        {
            "@odata.id": "/redfish/v1/Systems/1/Memory/proc1dimm1/",
            "DeviceLocator": "PROC 1 DIMM 1",
            "MemoryType": "NVDIMM_P",
            "Oem": {
                "Hpe": {
                    "@odata.context": "/redfish/v1/$metadata#HpeMemoryExt.HpeMemoryExt",
                    "@odata.type": "#HpeMemoryExt.v2_1_0.HpeMemoryExt",
                    "ProductName": "Intel Optane Memory"
                }
            }
        },
        {
            "@odata.id": "/redfish/v1/Systems/1/Memory/proc1dimm2/",
            "DeviceLocator": "PROC 1 DIMM 2",
            "MemoryType": "DRAM",
            "Oem": {
                "Hpe": {
                    "@odata.context": "/redfish/v1/$metadata#HpeMemoryExt.HpeMemoryExt",
                    "@odata.type": "#HpeMemoryExt.v2_1_0.HpeMemoryExt",
                    "DIMMStatus": "NotPresent"
                }
            }
        }
    ]

    chunks = [
        {
            "InterleaveSets": [
                {
                    "Memory": {
                        "@odata.id": "/redfish/v1/Systems/1/Memory/proc1dimm1/"
                    }
                }
            ]
        },
        {
            "InterleaveSets": [
                {
                    "Memory": {
                        "@odata.id": "/redfish/v1/Systems/1/Memory/proc1dimm12/"
                    }
                }
            ]
        },
        {
            "InterleaveSets": [
                {
                    "Memory": {
                        "@odata.id": "/redfish/v1/Systems/1/Memory/proc2dimm1/"
                    }
                },
                {
                    "Memory": {
                        "@odata.id": "/redfish/v1/Systems/1/Memory/proc2dimm12/"
                    },
                }
            ]
        }
    ]

    @staticmethod
    @pytest.mark.parametrize('dimm_id_list, expected_output', [
        (['1@2', '2@3', '3@4'], ['PROC 1 DIMM 2', 'PROC 2 DIMM 3', 'PROC 3 DIMM 4']),
        (['2@12'], ['PROC 2 DIMM 12'])
    ])
    def test_parse_dimm_id(dimm_id_list, expected_output):
        """
        Test for parse_dimm_id()
        :param dimm_id_list: DIMM IDs in the 'X@Y' format
        :param expected_output: expected output of the test
        """
        result = PmemHelpers.PmemHelpers.parse_dimm_id(dimm_id_list)
        assert result == expected_output

    @staticmethod
    @pytest.mark.parametrize("memory_members, expected_output", [
        (mem_members, ([{'MemoryType': 'NVDIMM_P',
                         '@odata.id': '/redfish/v1/Systems/1/Memory/proc1dimm1/',
                         'Oem': {
                             'Hpe':
                                 {'@odata.type': '#HpeMemoryExt.v2_1_0.HpeMemoryExt',
                                  '@odata.context': '/redfish/v1/$metadata#HpeMemoryExt.'
                                                    'HpeMemoryExt',
                                  'ProductName': 'Intel Optane Memory'}
                         },
                         'DeviceLocator': 'PROC 1 DIMM 1'}],
                       set(['PROC 1 DIMM 1']))),
        ([mem_members[0]], ([{'MemoryType': 'NVDIMM_P',
                              '@odata.id': '/redfish/v1/Systems/1/Memory/proc1dimm1/',
                              'Oem': {
                                  'Hpe':
                                      {'@odata.type': '#HpeMemoryExt.v2_1_0.HpeMemoryExt',
                                       '@odata.context': '/redfish/v1/$metadata#HpeMemoryExt.'
                                                         'HpeMemoryExt',
                                       'ProductName': 'Intel Optane Memory'}
                              },
                              'DeviceLocator': 'PROC 1 DIMM 1'}],
                            set(['PROC 1 DIMM 1']))),
        ([mem_members[1]], ([], set([]))),
        ([], ([], set([])))
    ])
    def test_get_pmem_members(memory_members, expected_output):
        """
        Test for get_pmem_members()
        :param memory_members: members of memory collection resource
        :param expected_output: expected output of the test
        """
        result = PmemHelpers.PmemHelpers.get_pmem_members(memory_members)
        assert result == expected_output

    @staticmethod
    @pytest.mark.parametrize("data_id, chunks, expected_output", [
        ("/redfish/v1/Systems/1/Memory/proc1dimm1/", chunks, False),
        ("/redfish/v1/Systems/1/Memory/proc2dimm1/", chunks, True),
        ("/redfish/v1/Systems/1/Memory/proc2dimm1/", [], None),
        ("/redfish/v1/Systems/1/Memory/proc1dimm4/", chunks, None)])
    def test_is_interleave_set(data_id, chunks, expected_output):
        """
        Test for is_interleave_set()
        :param data_id: @odata.id of the memory to be checked
        :param chunks: list of chunks
        :param expected_output: expected output of the test
        """
        result = PmemHelpers.PmemHelpers.is_interleave_set(data_id, chunks)
        assert result == expected_output

    @staticmethod
    @pytest.mark.parametrize('location_list, expected_output', [
        (['PROC X DIMM Y'], ('X@Y', 'PROC X')),
        (['PROC A DIMM B', 'PROC A DIMM C', 'PROC A DIMM D'], ('A@B, A@C, A@D', 'PROC A'))
    ])
    def test_location_format_converter(location_list, expected_output):
        """
        Test for location_format_converter()
        :param location_list: list of locations of the format 'PROC X DIMM Y'
        :param expected_output: expected output of the test
        """
        result = PmemHelpers.PmemHelpers.location_format_converter(location_list)
        assert result == expected_output

    @staticmethod
    @pytest.mark.parametrize("id1, id2, expected_output", [
        ("/redfish/v1/Systems/1/Memory/proc1dimm1", "/redfish/v1/Systems/1/Memory/proc1dimm1",
         True),
        ("/redfish/v1/Systems/1/Memory/proc1dimm1/", "/redfish/v1/Systems/1/Memory/proc1dimm1",
         True),
        ("/redfish/v1/Systems/1/Memory/proc1dimm1", "/redfish/v1/Systems/1/Memory/proc1dimm1/",
         True),
        ("/redfish/v1/Systems/1/Memory/proc1dimm1/", "/redfish/v1/Systems/1/Memory/proc1dimm1/",
         True),
        ("/redfish/v1/Systems/1/Memory/proc1dimm2/", "/redfish/v1/Systems/1/Memory/proc1dimm1/",
         False)
    ])
    def test_compare_id(id1, id2, expected_output):
        """
        Test for compare_id()
        :param id1: first id to be compared
        :param id2: second id to be compared
        :param expected_output:
        :return: expected output of the test
        """
        result = PmemHelpers.PmemHelpers.compare_id(id1, id2)
        assert result == expected_output



"""Module containing functional tests for the config flag"""
from __future__ import absolute_import

import sys

from rdmc_helper import ReturnCodes


class TestConfigFlag(object):
    """Functional Tests for the --config | -C flag"""

    @staticmethod
    def test_with_no_dimm_id(ilorest, dl360_large):
        """Test for --config flag with no DIMM IDs specified"""
        url, username, password = dl360_large.config
        with ilorest.session(url, username, password):
            results = ilorest.run(['showdcpmm', '--config'])
            sys.stdout.write("results.stdout: ")
            sys.stdout.write(results.stdout)
            sys.stdout.write("results.stderr: ")
            sys.stdout.write(results.stderr)
            assert results.retcode == ReturnCodes.SUCCESS

    @staticmethod
    def test_with_valid_dimm_id(ilorest, dl360_large):
        """Test for --config flag with valid DIMM IDs specified"""
        url, username, password = dl360_large.config
        with ilorest.session(url, username, password):
            results = ilorest.run(['showdcpmm', '--config', '--dimm', '1@2,2@3'])
            sys.stdout.write("results.stdout: ")
            sys.stdout.write(results.stdout)
            sys.stdout.write("results.stderr: ")
            sys.stdout.write(results.stderr)
            assert results.retcode == ReturnCodes.SUCCESS

    @staticmethod
    def test_with_non_pmem_dimm_id(ilorest, dl360_large):
        """Test for --config flag with valid non pmem DIMM IDs specified"""
        url, username, password = dl360_large.config
        with ilorest.session(url, username, password):
            results = ilorest.run(['showdcpmm', '--config', '--dimm', '1@5'])
            sys.stdout.write("results.stdout: ")
            sys.stdout.write(results.stdout)
            sys.stdout.write("results.stderr: ")
            sys.stdout.write(results.stderr)
            assert results.retcode == ReturnCodes.INVALID_COMMAND_LINE_ERROR

    @staticmethod
    def test_with_json_format_1(ilorest, dl360_large):
        """Test for --config flag with --json flag"""
        url, username, password = dl360_large.config
        with ilorest.session(url, username, password):
            results = ilorest.run(['showdcpmm', '--config', '--dimm', '1@2,2@3', '-j'])
            sys.stdout.write("results.stdout: ")
            sys.stdout.write(results.stdout)
            sys.stdout.write("results.stderr: ")
            sys.stdout.write(results.stderr)
            assert results.retcode == ReturnCodes.SUCCESS

    @staticmethod
    def test_with_json_format_2(ilorest, dl360_large):
        """Test for --config flag with --json flag"""
        url, username, password = dl360_large.config
        with ilorest.session(url, username, password):
            results = ilorest.run(['showdcpmm', '--config', '-j'])
            sys.stdout.write("results.stdout: ")
            sys.stdout.write(results.stdout)
            sys.stdout.write("results.stderr: ")
            sys.stdout.write(results.stderr)
            assert results.retcode == ReturnCodes.SUCCESS

    @staticmethod
    def test_with_invalid_member_dimm_id(ilorest, dl360_large):
        """Test for --config flag with invalid DIMM IDs specified"""
        url, username, password = dl360_large.config
        with ilorest.session(url, username, password):
            results = ilorest.run(['showdcpmm', '--config', '--dimm', '8@15'])
            sys.stdout.write("results.stderr: ")
            sys.stdout.write(results.stderr)
            assert results.retcode == ReturnCodes.INVALID_COMMAND_LINE_ERROR

    @staticmethod
    def test_with_invalid_dimm_id(ilorest):
        """Test for --config flag with invalid DIMM IDs specified"""
        results = ilorest.run(['showdcpmm', '--config', '--dimm', '132,243,xyz'])
        sys.stdout.write("results.stderr: ")
        sys.stdout.write(results.stderr)
        assert results.retcode == ReturnCodes.INVALID_COMMAND_LINE_ERROR



"""Module containing functional tests for the summary flag"""
from __future__ import absolute_import

import sys

from rdmc_helper import ReturnCodes


class TestSummaryFlag(object):
    """Functional Tests for the --summary | -M flag"""

    @staticmethod
    def test_with_no_attributes(ilorest, dl360_large):
        """Test for --summary flag with no attributes specified"""
        url, username, password = dl360_large.config
        with ilorest.session(url, username, password):
            results = ilorest.run(['showdcpmm', '--summary'])
            sys.stdout.write("results.stdout: ")
            sys.stdout.write(results.stdout)
            sys.stdout.write("results.stderr: ")
            sys.stdout.write(results.stderr)
            assert results.retcode == ReturnCodes.SUCCESS

    @staticmethod
    def test_with_dimm_flag(ilorest):
        """Test for --summary flag with dimm flag specified"""
        results = ilorest.run(['showdcpmm', '--summary', '--dimm', '1@2'])
        sys.stdout.write("results.stdout: ")
        sys.stdout.write(results.stdout)
        sys.stdout.write("results.stderr: ")
        sys.stdout.write(results.stderr)
        assert results.retcode == ReturnCodes.INVALID_COMMAND_LINE_ERROR

    @staticmethod
    def test_with_json_flag(ilorest, dl360_large):
        """Test for --summary flag with --json flag specified"""
        url, username, password = dl360_large.config
        with ilorest.session(url, username, password):
            results = ilorest.run(['showdcpmm', '--summary', '--json'])
            sys.stdout.write("results.stdout: ")
            sys.stdout.write(results.stdout)
            sys.stdout.write("results.stderr: ")
            sys.stdout.write(results.stderr)
            assert results.retcode == ReturnCodes.SUCCESS



###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
""" Save Command for RDMC """

import json
from collections import OrderedDict

import redfish.ris

try:
    from rdmc_helper import (
        Encryption,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        InvalidFileFormattingError,
        ReturnCodes,
        iLORisCorruptionError,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        Encryption,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        InvalidFileFormattingError,
        ReturnCodes,
        iLORisCorruptionError,
    )

# default file name
__filename__ = "ilorest.json"


class SaveCommand:
    """Constructor"""

    def __init__(self):
        self.ident = {
            "name": "save",
            "usage": None,
            "description": "Run to save a selected type to a file"
            "\n\texample: save --selector HpBios.\n\n\tChange the default "
            "output filename\n\texample: save --selector HpBios. -f "
            "output.json\n\n\tTo save multiple types in one file\n\texample: "
            "save --multisave Bios.,ComputerSystem.",
            "summary": "Saves the selected type's settings to a file.",
            "aliases": [],
            "auxcommands": ["SelectCommand"],
        }
        self.filename = __filename__
        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()

    def run(self, line, help_disp=False):
        """Main save worker function

        :param line: command line input
        :type line: string.
        """
        if help_disp:
            self.parser.print_help()
            return ReturnCodes.SUCCESS
        try:
            (options, args) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")

        if options.multisave:
            options.selector = ""
        if (
            hasattr(options, "selector")
            and "hpbaseconfigs" not in options.selector.lower()
            and "hpebaseconfigs" not in options.selector.lower()
            and "bios" not in options.selector.lower()
        ):
            self.savevalidation(options)
        else:
            # filename validations and checks
            self.cmdbase.login_validation(self, options)
            self.filename = None
            if options.filename and len(options.filename) > 1:
                raise InvalidCommandLineError("Save command doesn't support multiple filenames.")
            elif options.filename:
                self.filename = options.filename[0]
            elif self.rdmc.config:
                if self.rdmc.config.defaultsavefilename:
                    self.filename = self.rdmc.config.defaultsavefilename

            if not self.filename:
                self.filename = __filename__

        # if args:
        #  raise InvalidCommandLineError("Save command takes no arguments.")

        self.rdmc.ui.printer("Saving configuration...\n")
        if options.filter:
            try:
                if (str(options.filter)[0] == str(options.filter)[-1]) and str(options.filter).startswith(("'", '"')):
                    options.filter = options.filter[1:-1]

                (sel, val) = options.filter.split("=")
                sel = sel.strip()
                val = val.strip()
            except:
                raise InvalidCommandLineError("Invalid filter" " parameter format [filter_attribute]=[filter_value]")

            instances = self.rdmc.app.select(
                selector=self.rdmc.app.selector,
                fltrvals=(sel, val),
                path_refresh=True,
            )
            contents = self.saveworkerfunction(options, instances=instances)
        else:
            contents = self.saveworkerfunction(options)

        if options.multisave:
            for select in options.multisave:
                try:
                    self.auxcommands["select"].run(select)
                except:
                    pass
                contents += self.saveworkerfunction(options)

        if not contents:
            raise redfish.ris.NothingSelectedError
        else:
            # TODO: Maybe add this to the command. Not sure we use it elsewhere in the lib
            contents = self.add_save_file_header(contents)

        if options.encryption:
            with open(self.filename, "wb") as outfile:
                outfile.write(
                    Encryption().encrypt_file(
                        json.dumps(contents, indent=2, cls=redfish.ris.JSONEncoder),
                        options.encryption,
                    )
                )
        else:
            with open(self.filename, "w") as outfile:
                outfile.write(json.dumps(contents, indent=2, cls=redfish.ris.JSONEncoder, sort_keys=True))
        self.rdmc.ui.printer("Configuration saved to: %s\n" % self.filename)

        self.cmdbase.logout_routine(self, options)

        # Return code
        return ReturnCodes.SUCCESS

    def saveworkerfunction(self, options, instances=None):
        """Returns the currently selected type for saving

        :param instances: list of instances from select to save
        :type instances: list.
        """
        try:
            if (
                "bios" in options.selector.lower()
                or "hpbaseconfigs" in options.selector.lower()
                or "hpebaseconfigs" in options.selector.lower()
            ) and (options.multisave is None or options.multisave == ""):
                raise KeyError
            content = self.rdmc.app.getprops(insts=instances)
            try:
                contents = [{val[self.rdmc.app.typepath.defs.hrefstring]: val} for val in content]
            except KeyError:
                try:
                    contents = [{val["links"]["self"][self.rdmc.app.typepath.defs.hrefstring]: val} for val in content]
                except KeyError:
                    raise iLORisCorruptionError(
                        "iLO Database seems to be corrupted. Please check. Reboot the server to " "restore\n"
                    )
        except Exception:
            config_path = None
            contents = list()
            if options.selector.lower() == "bios." or "bios.v" in options.selector.lower():
                config_path = ["/redfish/v1/systems/1/bios/settings/"]
            elif options.selector.lower() == "bios":
                config_path = ["/redfish/v1/systems/1/bios/settings/", "/redfish/v1/systems/1/bios/mappings/"]
            elif "hpbaseconfigs" in options.selector.lower() or "hpebaseconfigs" in options.selector.lower():
                config_path = [
                    "/redfish/v1/systems/1/bios/oem/hpe/baseconfigs/",
                    "/redfish/v1/systems/1/bios/oem/hpe/nvmeof/baseconfigs/",
                    "/redfish/v1/systems/1/bios/oem/hpe/iscsi/baseconfigs/",
                    "/redfish/v1/systems/1/bios/oem/hpe/tlsconfig/baseconfigs/",
                    "/redfish/v1/systems/1/bios/oem/hpe/serverconfiglock/baseconfigs/",
                    "/redfish/v1/systems/1/bios/oem/hpe/kmsconfig/baseconfigs/",
                    "/redfish/v1/systems/1/bios/oem/hpe/boot/baseconfigs/",
                    "/redfish/v1/systems/1/bios/baseconfigs/",
                    "/redfish/v1/systems/1/bios/nvmeof/baseconfigs/",
                    "/redfish/v1/systems/1/bios/iscsi/baseconfigs/",
                    "/redfish/v1/systems/1/bios/tlsconfig/baseconfigs/",
                    "/redfish/v1/systems/1/bios/serverconfiglock/baseconfigs/",
                    "/redfish/v1/systems/1/bios/kmsconfig/baseconfigs/",
                    "/redfish/v1/systems/1/bios/boot/baseconfigs/",
                ]
            result = None
            for b in config_path:
                try:
                    result = self.rdmc.app.get_handler(b, silent=True, service=True)
                except:
                    pass
                if result is not None:
                    if result.status == 200:
                        d = {b: result.dict}
                        contents.append(d)

        type_string = self.rdmc.app.typepath.defs.typestring

        templist = list()
        srnum_list = list()

        for content in contents:
            typeselector = None
            pathselector = None

            for path, values in content.items():
                # if "Managers/1/EthernetInterfaces/1" not in path:
                for dictentry in list(values.keys()):
                    if dictentry == type_string:
                        typeselector = values[dictentry]
                        pathselector = path
                        del values[dictentry]
                    if dictentry in [
                        "IPv4Addresses",
                        "IPv6Addresses",
                        "IPv6AddressPolicyTable",
                        "MACAddress",
                        "StaticNameServers",
                        # "AutoNeg",
                        # "FullDuplex",
                        "SpeedMbps",
                    ]:
                        del values[dictentry]
                    if dictentry in [
                        "IPv6StaticAddresses",
                        "IPv6StaticDefaultGateways",
                        "IPv4StaticAddresses",
                    ]:
                        if values[dictentry]:
                            del values[dictentry]
                    if dictentry in ["IPv4", "IPv6", "DHCPv6", "DHCPv4"]:
                        if "DNSServers" in values[dictentry]:
                            del values[dictentry]["DNSServers"]
                            del values["Oem"]["Hpe"][dictentry]["DNSServers"]
                        if "UseDNSServers" in values[dictentry]:
                            del values[dictentry]["UseDNSServers"]
                            del values["Oem"]["Hpe"][dictentry]["UseDNSServers"]

                if values:
                    skip = False
                    if "SerialNumber" in values and values["SerialNumber"] is not None and values["SerialNumber"] != "":
                        if values["SerialNumber"] in srnum_list:
                            skip = True
                        else:
                            srnum_list.append(values["SerialNumber"])
                    tempcontents = dict()

                    if not skip:
                        if typeselector and pathselector:
                            tempcontents[typeselector] = {pathselector: values}
                        else:
                            raise InvalidFileFormattingError("Missing path or selector in input file.")

                templist.append(tempcontents)

        return templist

    def nested_sort(self, data):
        """Helper function to sort all dictionary key:value pairs

        :param data: dictionary to sort
        :type data: dict.
        """

        for key, value in data.items():
            if isinstance(value, dict):
                data[key] = self.nested_sort(value)

        data = OrderedDict(sorted(list(data.items()), key=lambda x: x[0]))

        return data

    def savevalidation(self, options):
        """Save method validation function

        :param options: command line options
        :type options: list.
        """

        if options.multisave:
            options.multisave = options.multisave.replace('"', "").replace("'", "")
            options.multisave = options.multisave.replace(" ", "").split(",")
            if not len(options.multisave) >= 1:
                raise InvalidCommandLineError("Invalid number of types in multisave option.")
            options.selector = options.multisave[0]
            options.multisave = options.multisave[1:]

        self.cmdbase.login_select_validation(self, options)

        # filename validations and checks
        self.filename = None

        if options.filename and len(options.filename) > 1:
            raise InvalidCommandLineError("Save command doesn't support multiple filenames.")
        elif options.filename:
            self.filename = options.filename[0]
        elif self.rdmc.config:
            if self.rdmc.config.defaultsavefilename:
                self.filename = self.rdmc.config.defaultsavefilename

        if not self.filename:
            self.filename = __filename__

    def add_save_file_header(self, contents):
        """Helper function to retrieve the comments for save file

        :param contents: current save contents
        :type contents: list.
        """
        templist = list()

        headers = self.rdmc.app.create_save_header()
        templist.append(headers)

        for content in contents:
            templist.append(content)

        return templist

    def definearguments(self, customparser):
        """Wrapper function for new command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)
        customparser.add_argument(
            "-f",
            "--filename",
            dest="filename",
            help="Use this flag if you wish to use a different filename than the default one. "
            "The default filename is %s." % __filename__,
            action="append",
            default=None,
        )

        customparser.add_argument(
            "--selector",
            dest="selector",
            help="Optionally include this flag to select a type to run the current command on. "
            "Use this flag when you wish to select a type without entering another command, "
            "or if you wish to work with a type that is different from the one currently "
            "selected.",
            default=None,
            # required=True,
        )
        customparser.add_argument(
            "--multisave",
            dest="multisave",
            help="Optionally include this flag to save multiple types to a single file. "
            "Overrides the currently selected type.\n\t Usage: --multisave type1.,type2.,type3.",
            default="",
        )
        customparser.add_argument(
            "--filter",
            dest="filter",
            help="Optionally set a filter value for a filter attribute. This uses the provided "
            "filter for the currently selected type. Note: Use this flag to narrow down your "
            "results. For example, selecting a common type might return multiple objects that "
            "are all of that type. If you want to modify the properties of only one of those "
            "objects, use the filter flag to narrow down results based on properties."
            "\n\t Usage: --filter [ATTRIBUTE]=[VALUE]",
            default=None,
        )
        customparser.add_argument(
            "-j",
            "--json",
            dest="json",
            action="store_true",
            help="Optionally include this flag if you wish to change the displayed output to "
            "JSON format. Preserving the JSON data structure makes the information easier to "
            "parse.",
            default=False,
        )
        customparser.add_argument(
            "--encryption",
            dest="encryption",
            help="Optionally include this flag to encrypt/decrypt a file using the key provided.",
            default=None,
        )



"""Module to test the DisplayHelpers Class"""
from __future__ import absolute_import

import ast
import sys

import pytest

# For cross-compatibility between Python 2 and Python 3
from six import StringIO

from .conftest import DisplayHelpers


class TestDisplayHelpers(object):
    """
    Class to test methods from the DisplayHelpers class
    """
    output_format = DisplayHelpers.OutputFormats

    @staticmethod
    def capture_stdout(func, data, out_format, prop_id=None):
        """
        Function to capture the standard output stream
        """
        old_stdout = sys.stdout
        capturer = StringIO()
        sys.stdout = capturer
        func(data, out_format, prop_id)
        sys.stdout = old_stdout
        output = capturer.getvalue()
        return output

    @pytest.mark.parametrize('table_data, expected_output', [
        # Tests with empty input
        (None, ""),
        ([], ""),
        # Single object test
        (["ItemId:ID1\nProperty1:Value1"], "ItemId Property1 ID1 Value1"),
        # Test to check truncation of string
        (["ItemId:ID1\nProp2:Value of property2\nProp1:Value1"],
         "ItemId Prop2 Prop1 ID1 Value of.. Value1"),
        # Multiple objects test
        (["ItemId:ID1\nProperty1: ", "ItemId:ID2\nProperty1:LongPropertyValue"],
         "ItemId Property1 ID1 ID2 LongProp..")
    ])
    def test_printtable_withwidth(self, table_data, expected_output):
        """
        Test for printtable with a custom specified width
        """
        display_helpers = DisplayHelpers.DisplayHelpers(10)
        out = TestDisplayHelpers.capture_stdout(
            display_helpers.display_data, table_data, self.output_format.table, None)
        assert [str(x) for x in out.split()] == expected_output.split()

    @pytest.mark.parametrize('table_data, expected_output', [
        # Tests with empty input
        (None, ""),
        ([], ""),
        # Single object test
        (["ItemId:ID1\nProperty1:Value1"], "ItemId Property1 ID1 Value1"),
        # Test to check truncation of string
        (["ItemId:ID1\nProp2:Value of property2\nProp1:Value1"],
         "ItemId Prop2 Prop1 ID1 Value of property2 Value1"),
        # Converting redfish property to friendly name test
        (["Property2:Value2\nFWVersion:2.1"], " Property2 FWVersion Value2 2.1"),
        # Multiple objects test
        (["ItemId:ID1\nProperty1: ", "ItemId:ID2\nProperty1:LongPropertyValue"],
         "ItemId Property1 ID1 ID2 LongPropertyValue")
    ])
    def test_printtable(self, table_data, expected_output):
        """
        Test for printtable with the default width
        """
        display_helpers = DisplayHelpers.DisplayHelpers()
        out = TestDisplayHelpers.capture_stdout(
            display_helpers.display_data, table_data, self.output_format.table, None)
        assert out.split() == expected_output.split()

    @pytest.mark.parametrize('table_data, expected_output', [
        # Tests with empty input
        (None, ""),
        ([], ""),
        # Single object test
        (["ItemId:ID1\nProperty1:Value1"],
         "--- 1 ---\nItemId: ID1\nProperty1: Value1"),
        # Converting redfish property to friendly name test
        (["FWVersion:2.1\nItemId:ID1"], "--- 1 ---\nFWVersion: 2.1\nItemId: ID1"),
        # Multiple objects test
        (["ItemId:ID1\nProperty1:", "ItemId:ID2\nProperty1:Value1"],
         "--- 1 ---\nItemId: ID1\nProperty1: \n\n--- 2 ---\nItemId: ID2\nProperty1: Value1")
    ])
    def test_printlist(self, table_data, expected_output):
        """
        Test for the printlist function that produces output in list form
        """
        display_helpers = DisplayHelpers.DisplayHelpers()
        out = TestDisplayHelpers.capture_stdout(
            display_helpers.display_data, table_data, self.output_format.list, None)
        assert out.strip() == expected_output.strip()

    @pytest.mark.parametrize('table_data, expected_output', [
        # Tests with empty input
        (None, ""),
        ([], ""),
        # Single object test
        (["ItemId:ID1\nProperty1:Value1"],
         "--- ItemId: ID1 ---\nProperty1: Value1"),
        # Converting redfish property to friendly name test
        (["ItemId:ID1\nFWVersion:2.1"], "--- ItemId: ID1 ---\nFWVersion: 2.1"),
        # Multiple objects test
        (["ItemId:ID1\nProperty1:", "ItemId:ID2\nProperty1:Value1"],
         "--- ItemId: ID1 ---\nProperty1: \n\n--- ItemId: ID2 ---\nProperty1: Value1")
    ])
    def test_printlist_withid(self, table_data, expected_output):
        """
        Test for the printlist function that produces output in list form with custom ID
        """
        display_helpers = DisplayHelpers.DisplayHelpers()
        out = TestDisplayHelpers.capture_stdout(
            display_helpers.display_data, table_data, self.output_format.list, "ItemId")
        assert out.strip() == expected_output.strip()

    @pytest.mark.parametrize('table_data, expected_output', [
        # Tests with empty input
        (None, ""),
        ([], ""),
        # Single object test
        ([{"ItemId": "ID1", "Property1": "Value1"}],
         [{"ItemId": "ID1", "Property1": "Value1"}]),
        # Converting redfish property to friendly name test
        ([{"FWVersion": 2.1, "Property2": "Value2"}],
         [{"Property2": "Value2", "FWVersion": 2.1}]),
        # Multiple objects test
        ([{"ItemId": "ID1", "Property1": ""}, {"ItemId": "ID2", "Property1": "LongPropertyValue"}],
         [{"ItemId": "ID1", "Property1": ""}, {"ItemId": "ID2", "Property1": "LongPropertyValue"}])
    ])
    def test_printjson(self, table_data, expected_output):
        """
        Test for the printjson function that produces output in json form
        """
        display_helpers = DisplayHelpers.DisplayHelpers()
        out = TestDisplayHelpers.capture_stdout(
            display_helpers.display_data, table_data, self.output_format.json, None)
        if expected_output == "":
            assert out == expected_output
        else:
            out = ast.literal_eval("".join(out.split()))
            assert out == expected_output

    @staticmethod
    @pytest.mark.parametrize('string, maxlen, expected_output', [
        ("Banana", 10, "Banana"),
        ("Hot Tamale", 10, "Hot Tamale"),
        ("Antidisestablishmentarianism", 10, "Antidise.."),
        ("hippopotomonstrosesquipedaliophobia", 18, "hippopotomonstro.."),
        (None, 10, "")
    ])
    def test_truncate_lengthy(string, maxlen, expected_output):
        """
        Test for the truncate_lengthy method
        """
        display_helpers = DisplayHelpers.DisplayHelpers()
        out = display_helpers.truncate_lengthy(string, maxlen)
        assert out == expected_output



###
# Copyright 2017 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""This is the helper module for RDMC"""

# ---------Imports---------
from __future__ import unicode_literals

import os
import sys
from argparse import (
    SUPPRESS,
    Action,
    ArgumentParser,
    RawDescriptionHelpFormatter,
    _ArgumentGroup,
)
from builtins import bytes, int, object, str, super

from redfish.ris import NothingSelectedError

try:
    import cliutils
except ImportError:
    from ilorest import cliutils

try:
    import versioning
except ImportError:
    from ilorest import versioning
try:
    import rdmc_helper
except ImportError:
    from ilorest import rdmc_helper

try:
    from rdmc_helper import InvalidCommandLineErrorOPTS
except ImportError:
    from ilorest.rdmc_helper import InvalidCommandLineErrorOPTS


# from extensions.COMMANDS import LoginCommand, SelectCommand

# ---------End of imports---------


class _Verbosity(Action):
    def __init__(self, option_strings, dest, nargs, **kwargs):
        super(_Verbosity, self).__init__(option_strings, dest, nargs, **kwargs)

    def __call__(self, parser, namespace, values, option_strings):
        try:
            if values:
                tmp = next(iter(values))
                if tmp.isdigit():
                    namespace.verbose = int(tmp)
                else:
                    namespace.verbose = len(tmp) + 1
                return
            namespace.verbose = 1
        except:
            raise InvalidCommandLineErrorOPTS("Invalid verbosity selection ('-v').")


class CommandBase(object):
    """Abstract base class for all Command objects.

    This class is used to build complex command line programs
    """

    def __init__(self, name, usage, summary, aliases=None, argparser=None):
        self.name = name
        self.summary = summary
        self.aliases = aliases
        self.config_required = True  # does the command access config data

        if argparser is None:
            self.parser = ArgumentParser()
        else:
            self.parser = argparser

        self.parser.usage = usage
        # TODO: See if we can remove this or stop it from opening a subprocess all the time
        # self._cli = cliutils.CLI()

    def run(self, line, help_disp=False):
        """Called to actually perform the work.

        Override this method in your derived class.  This is where your program
        actually does work.
        """
        pass


class RdmcCommandBase(CommandBase):
    """Base class for rdmc commands which includes some common helper
    methods.
    """

    def __init__(self, name, usage, summary, aliases, argparser=None, **kwargs):
        """Constructor"""
        CommandBase.__init__(
            self, name=name, usage=usage, summary=summary, aliases=aliases, argparser=argparser, **kwargs
        )
        self.json = False
        self.cache = False
        self.nologo = False
        self.toolbar = False

    def login_validation(self, cmdinstance, options, skipbuild=False):
        """Combined validation function to login and select with other commands. Not for use with
        login or select commands themselves. Make sure your command imports options from
        add_login_arguments_group or there will be errors.

        :param cmdinstance: the command object instance
        :type cmdinstance: list.
        :param options: command line options
        :type options: list.
        :param skipbuild: flag to only login and skip monolith build
        :type skipbuild: bool.
        """

        logobj = cmdinstance.rdmc.load_command(cmdinstance.rdmc.search_commands("LoginCommand"))
        inputline = list()
        client = None
        loggedin = False

        if hasattr(options, "json") and cmdinstance.rdmc.config.format.lower() == "json":
            options.json = True

        try:
            client = cmdinstance.rdmc.app.current_client
        except:
            if options.user or options.password or options.url or options.force_vnic:
                if options.url:
                    inputline.extend([options.url])
                if options.user:
                    if options.encode:
                        options.user = rdmc_helper.Encryption.decode_credentials(options.user)
                        if isinstance(options.user, bytes):
                            options.user = options.user.decode("utf-8")
                    inputline.extend(["-u", options.user])
                if options.password:
                    if options.encode:
                        options.password = rdmc_helper.Encryption.decode_credentials(options.password)
                        if isinstance(options.password, bytes):
                            options.password = options.password.decode("utf-8")
                    inputline.extend(["-p", options.password])
                if options.force_vnic:
                    inputline.extend(["--force_vnic"])
                if getattr(options, "https_cert", False):
                    inputline.extend(["--https", options.https_cert])
                if getattr(options, "user_certificate", False):
                    inputline.extend(["--usercert", options.user_certificate])
                if getattr(options, "user_root_ca_key", False):
                    inputline.extend(["--userkey", options.user_root_ca_key])
                if getattr(options, "user_root_ca_password", False):
                    inputline.extend(["--userpassphrase", options.user_root_ca_password])
            else:
                if cmdinstance.rdmc.config.url:
                    inputline.extend([cmdinstance.rdmc.config.url])
                if cmdinstance.rdmc.config.username:
                    inputline.extend(["-u", cmdinstance.rdmc.config.username])
                if cmdinstance.rdmc.config.password:
                    inputline.extend(["-p", cmdinstance.rdmc.config.password])
                if cmdinstance.rdmc.config.ssl_cert:
                    inputline.extend(["--https", cmdinstance.rdmc.config.ssl_cert])
                if getattr(options, "user_certificate", False):
                    inputline.extend(["--usercert", options.user_certificate])
                if getattr(options, "user_root_ca_key", False):
                    inputline.extend(["--userkey", options.user_root_ca_key])
                if getattr(options, "user_root_ca_password", False):
                    inputline.extend(["--userpassphrase", options.user_root_ca_password])
            if options.includelogs:
                inputline.extend(["--includelogs"])
            if options.path:
                inputline.extend(["--path", options.path])

            if getattr(options, "biospassword", False):
                inputline.extend(["--biospassword", options.biospassword])
            if getattr(options, "sessionid", False):
                inputline.extend(["--sessionid", options.sessionid])

            logobj.loginfunction(inputline, skipbuild=skipbuild)
            loggedin = True

        if not (loggedin or client or options.url or cmdinstance.rdmc.app.typepath.url):
            message = "Local login initiated...\n"
            if cmdinstance.rdmc.opts.verbose:
                sys.stdout.write(message)
            else:
                rdmc_helper.LOGGER.info(message)

        if not (loggedin or client):
            logobj.loginfunction(inputline, skipbuild=skipbuild)

    def login_select_validation(self, cmdinstance, options, skipbuild=False):
        """Combined validation function to login and select with other commands. Not for use with
        login or select commands themselves. Make sure your command imports options from
        add_login_arguments_group or there will be errors.

        :param cmdinstance: the command object instance
        :type cmdinstance: list.
        :param options: command line options
        :type options: list.
        :param skipbuild: flag to only login and skip monolith build
        :type skipbuild: bool.
        """
        inputline = list()

        self.login_validation(cmdinstance, options, skipbuild=skipbuild)
        logobj = cmdinstance.rdmc.load_command(cmdinstance.rdmc.search_commands("LoginCommand"))
        selobj = cmdinstance.rdmc.load_command(cmdinstance.rdmc.search_commands("SelectCommand"))
        if hasattr(options, "selector") and options.selector:
            if inputline:
                inputline.extend(["--selector", options.selector])
                logobj.loginfunction(inputline)
            else:
                if getattr(options, "ref", False):
                    inputline.extend(["--refresh"])

                inputline.extend([options.selector])
                selobj.selectfunction(inputline)
        elif hasattr(options, "selector"):
            try:
                inputline = list()
                selector = cmdinstance.rdmc.app.selector

                if hasattr(options, "ref") and options.ref:
                    inputline.extend(["--refresh"])

                if selector:
                    inputline.extend([selector])
                    selobj.selectfunction(inputline)
            except NothingSelectedError:
                raise NothingSelectedError

    def logout_routine(self, cmdinstance, options):
        """Routine to logout of a server automatically at the completion of a command.

        :param commandinstance: the command object instance
        :type commandinstance: list.
        :param options: command line options
        :type options: list.
        """

        logoutobj = cmdinstance.rdmc.load_command(cmdinstance.rdmc.search_commands("LogoutCommand"))

        if getattr(options, "logout", False):
            logoutobj.run("")

    def add_login_arguments_group(self, parser):
        """Adds login arguments to the passed parser

        :param parser: The parser to add the login option group to
        :type parser: ArgumentParser
        """
        group = parser.add_argument_group("LOGIN OPTIONS", "Options for logging in to a system.")
        group.add_argument("--url", dest="url", help="Use the provided iLO URL to login.", default=None)
        group.add_argument(
            "--sessionid",
            dest="sessionid",
            help="Use the provided sessionid to login.",
            default=None,
        )
        group.add_argument(
            "-u",
            "--user",
            dest="user",
            help="""If you are not logged in yet, including this flag along with the
password and URL flags can be used to login to a server in the same command.""",
            default=None,
        )
        group.add_argument(
            "-p",
            "--password",
            dest="password",
            help="""Use the provided iLO password to log in.""",
            default=None,
        )
        group.add_argument(
            "-o",
            "--otp",
            dest="login_otp",
            help="""Use the provided iLO OTP to log in.""",
            default=None,
        )
        group.add_argument(
            "--biospassword",
            dest="biospassword",
            help="""Select this flag to input a BIOS password. Include this
flag if second-level BIOS authentication is needed for the command to execute.
This option is only used on Gen 9 systems.""",
            default=None,
        )
        group.add_argument(
            "--https",
            dest="https_cert",
            help="""Use the provided CA bundle or SSL certificate with your login to
connect securely to the system in remote mode. This flag has no effect in local mode.""",
            default=None,
        )
        group.add_argument(
            "--usercert",
            dest="user_certificate",
            type=str,
            help="""Specify a user certificate file path for certificate based authentication
with iLO.\n**NOTE**: Inclusion of this argument will force certficate based
authentication. A root user certificate authority key or bundle will be required.""",
            default=None,
        )
        group.add_argument(
            "--userkey",
            dest="user_root_ca_key",
            type=str,
            help="""Specify a user root ca key file path for certificate based certificate
authentication with iLO. **NOTE 1**: Inclusion of this argument will force certficate based
authentication. A root user certificate authority key or bundle will be required.
**NOTE 2**: Inclusion of this argument will force certificate based authentication.
A user certificate will be required.
**NOTE 3**: A user will be prompted for a password if the root certificate authority key
is encrypted and \'-certpass/--userrootcapassword\' is omitted.""",
            default=None,
        )
        group.add_argument(
            "--userpassphrase",
            dest="user_root_ca_password",
            type=str,
            help="""Optionally specify a user root ca key file password for encrypted
user root certificate authority keys. **NOTE 1**: Inclusion of this argument will force
certficate based authentication. A root user certificate authority key or
bundle will be required. **NOTE 2**: The user will be prompted for a password
if the user root certificate authority key requires a password""",
            default=None,
        )
        # group.add_argument(
        #    '--certbundle',
        #    dest='ca_cert_bundle',
        #    type=str,
        #    help="""Specify a file path for the certificate authority bundle location
        # (local repository for certificate collection) **NOTE**: Providing a custom certificate
        # or root CA key will override the use of certificate bundles""",
        #            default=None)
        group.add_argument(
            "-e",
            "--enc",
            dest="encode",
            action="store_true",
            help=SUPPRESS,
            default=False,
        )
        group.add_argument(
            "--includelogs",
            dest="includelogs",
            action="store_true",
            help="Optionally include logs in the data retrieval process.",
            default=False,
        )
        group.add_argument(
            "--path",
            dest="path",
            help="""Optionally set a starting point for data collection during login.
If you do not specify a starting point, the default path will be /redfish/v1/.
Note: The path flag can only be specified at the time of login.
Warning: Only for advanced users, and generally not needed for normal operations.""",
            default=None,
        )
        group.add_argument(
            "--force_vnic",
            dest="force_vnic",
            action="store_true",
            help="Force login through iLO Virtual NIC. **NOTE** " "iLO 5 required",
            default=False,
        )
        group.add_argument(
            "--logout",
            dest="logout",
            action="store_true",
            help="Logout after the completion of the command.",
            default=None,
        )


class RdmcOptionParser(ArgumentParser):
    """Constructor"""

    def __init__(self):
        super().__init__(
            usage="%s [GLOBAL OPTIONS] [COMMAND] [COMMAND ARGUMENTS] " "[COMMAND OPTIONS]" % versioning.__shortname__,
            description="iLOrest is a command-line or interactive interface that allows users "
            "to manage Hewlett Packard Enterprise products that take advantage"
            " of RESTful APIs.\n\nIn order to view or manage a system you must"
            " first login. You can login using the login command or during "
            "execution of any other command.\nFrom here you can run any other "
            "commands. To learn more about specific commands, run iLOrest "
            "COMMAND -h.",
            epilog="Examples:\n\nThe following is the standard flow of command"
            "s to view system data.\n\tThe first example is each command "
            "run individually: \n\n\tilorest login\n\tilorest select Bios.\n\t"
            "ilorest get\n\n\tThe second is the list of all of the commands "
            "run at once. First locally, then remotely.\n\tilorest get "
            "--select Bios.\n\tilorest get --select Bios. --url <iLO IP> -u"
            " <iLO Username> -p <iLO Password>",
            formatter_class=RawDescriptionHelpFormatter,
        )
        globalgroup = _ArgumentGroup(self, "GLOBAL OPTIONS")

        self.add_argument(
            "--config",
            dest="config",
            help="Use the provided configuration file instead of the default one.",
            metavar="FILE",
        )

        config_dir_default = os.path.join(cliutils.get_user_config_dir(), ".%s" % versioning.__shortname__)
        self.add_argument(
            "--cache-dir",
            dest="config_dir",
            default=config_dir_default,
            help="Use the provided directory as the location to cache data"
            " (default location: %s)" % config_dir_default,
            metavar="PATH",
        )
        self.add_argument(
            "-v",
            "--verbose",
            dest="verbose",
            action="count",
            help="Display verbose information (with increasing level). '-v': Level 1, "
            "Logging, Stdout, Stderr. '-vv': Level 2, Extends Level 1 with slightly "
            "elaborated iLO and HTTP response message. '-vvv': Level3, Extends Level 2 "
            "with message id, validation class, message text with embedded args, and "
            "possible resolution/mitigation for iLO responses. Includes HTTP responses. "
            "**NOTE 1**: Some responses may only contain limited information from the source."
            "**NOTE 2**: Default level is 0.",
            default=0,
        )
        self.add_argument(
            "-d",
            "--debug",
            dest="debug",
            action="store_true",
            help="""Display debug information.""",
            default=False,
        )
        self.add_argument(
            "--logdir",
            dest="logdir",
            default=None,
            help="""Use the provided directory as the location for log file.""",
            metavar="PATH",
        )
        self.add_argument(
            "--nostdoutlog",
            dest="nostdoutlog",
            action="store_true",
            help="""Disable debug logs to stdout.""",
            default=False,
        )
        self.add_argument(
            "--nocache",
            dest="nocache",
            action="store_true",
            help="During execution the application will temporarily store data only in memory.",
            default=False,
        )
        self.add_argument(
            "--nologo",
            dest="nologo",
            action="store_true",
            help="""Include to block copyright and logo.""",
            default=False,
        )
        self.add_argument(
            "--toolbar",
            dest="toolbar",
            action="store_true",
            help="""Show toolbar at the bottom.""",
            default=False,
        )
        self.add_argument(
            "--notab",
            dest="notab",
            action="store_true",
            help="""Disable tab complete.""",
            default=False,
        )
        self.add_argument(
            "--redfish",
            dest="is_redfish",
            action="store_true",
            help="Use this flag if you wish to to enable "
            "Redfish only compliance. It is enabled by default "
            "in systems with iLO5 and above.",
            default=False,
        )
        self.add_argument(
            "--latestschema",
            dest="latestschema",
            action="store_true",
            help="Optionally use the latest schema instead of the one "
            "requested by the file. Note: May cause errors in some data "
            "retrieval due to difference in schema versions.",
            default=False,
        )
        self.add_argument(
            "--useproxy",
            dest="proxy",
            default=None,
            help="""Use the provided proxy for communication.""",
            metavar="URL",
        )
        self.add_argument(
            "--redirectconsole",
            dest="redirect",
            help="Optionally include this flag to redirect stdout/stderr console.",
            nargs="?",
            default=None,
            const=True,
            metavar="REDIRECT CONSOLE",
        )
        self.add_argument_group(globalgroup)



###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
""" Login Command for RDMC """

import getpass
import os
import socket
import redfish.ris

try:
    from rdmc_helper import (
        Encryption,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        PathUnavailableError,
        ReturnCodes,
        UsernamePasswordRequiredError,
    )
except ModuleNotFoundError:
    from ilorest.rdmc_helper import (
        ReturnCodes,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        PathUnavailableError,
        Encryption,
        UsernamePasswordRequiredError,
    )

from redfish.rest.v1 import ServerDownOrUnreachableError


class LoginCommand:
    """Constructor"""

    def __init__(self):
        self.ident = {
            "name": "login",
            "usage": None,
            "description": "To login remotely run using iLO url and iLO credentials"
            "\n\texample: login <iLO url/hostname> -u <iLO username> "
            "-p <iLO password>\n\n\tTo login on a local server run without "
            "arguments\n\texample: login"
            "\n\n\tTo login through VNIC run using --force_vnic and iLO credentials "
            "\n\texample: login --force_vnic -u <iLO username> -p <iLO password>"
            "\n\nLogin using OTP can be done in 2 ways."
            "\n\n\t To login implicitly, use the tag --wait_for_otp."
            "\n\t\texample: login -u <iLO username> -p <iLO password> --wait_for_otp"
            "\n\n\n\t To login explicitly, use the tag -o/--otp and enter OTP after."
            "\n\t\texample: login -u <iLO username> -p <iLO password> -o <iLO OTP>"
            "\n\n\tNOTE: A [URL] can be specified with "
            "an IPv4, IPv6, or hostname address.",
            "summary": "Connects to a server, establishes a secure session," " and discovers data from iLO.",
            "aliases": [],
            "auxcommands": ["LogoutCommand"],
            "cert_data": {},
        }
        self.cmdbase = None
        self.rdmc = None
        self.url = None
        self.username = None
        self.password = None
        self.sessionid = None
        self.biospassword = None
        self.auxcommands = dict()
        self.cert_data = dict()
        self.login_otp = None

    def run(self, line, help_disp=False):
        """wrapper function for main login function

        :param line: command line input
        :type line: string.
        :param help_disp: flag to determine to display or not
        :type help_disp: boolean
        """
        if help_disp:
            self.parser.print_help()
            return ReturnCodes.SUCCESS
        try:
            self.loginfunction(line)

            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS

            if not self.rdmc.app.monolith._visited_urls:
                self.auxcommands["logout"].run("")
                raise PathUnavailableError("The path specified by the --path flag is unavailable.")
        except Exception:
            raise

        # Return code
        return ReturnCodes.SUCCESS

    def perform_login(self, options, skipbuild, user_ca_cert_data):
        self.rdmc.app.login(
            username=self.username,
            password=self.password,
            sessionid=self.sessionid,
            base_url=self.url,
            path=options.path,
            skipbuild=skipbuild,
            includelogs=options.includelogs,
            biospassword=self.biospassword,
            is_redfish=self.rdmc.opts.is_redfish,
            proxy=self.rdmc.opts.proxy,
            user_ca_cert_data=user_ca_cert_data,
            json_out=self.rdmc.json,
            login_otp=self.login_otp,
            log_dir=self.rdmc.log_dir,
        )

    def loginfunction(self, line, skipbuild=None, json_out=False):
        """Main worker function for login class

        :param line: entered command line
        :type line: list.
        :param skipbuild: flag to determine if monolith should be build
        :type skipbuild: boolean.
        :param json_out: flag to determine if json output neededd
        :type skipbuild: boolean.
        """
        try:
            (options, args) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineError("Invalid command line arguments")

        self.loginvalidation(options, args)

        # if proxy server provided in command line as --useproxy, it will be used,
        # otherwise it will the environment variable setting.
        # else proxy will be set as None.
        if self.rdmc.opts.proxy:
            _ = self.rdmc.opts.proxy
        elif "https_proxy" in os.environ and os.environ["https_proxy"]:
            _ = os.environ["https_proxy"]
        elif "http_proxy" in os.environ and os.environ["http_proxy"]:
            _ = os.environ["http_proxy"]
        else:
            _ = self.rdmc.config.proxy

        no_bundle = False

        if getattr(options, "ca_cert_bundle", False):
            user_ca_cert_data = {"ca_certs": options.ca_cert_bundle}
        else:
            user_ca_cert_data = {}
        if getattr(options, "user_certificate", False):
            no_bundle = True
            user_ca_cert_data.update({"cert_file": options.user_certificate})
        if getattr(options, "user_root_ca_key", False):
            no_bundle = True
            user_ca_cert_data.update({"key_file": options.user_root_ca_key})
        if getattr(options, "user_root_ca_password", False):
            no_bundle = True
            user_ca_cert_data.update({"key_password": options.user_root_ca_password})

        if not no_bundle:
            if hasattr(user_ca_cert_data, "ca_certs"):
                user_ca_cert_data.pop("ca_certs")

        try:
            if getattr(options, "force_vnic", False):
                self.rdmc.ui.printer("\nAttempt to login with Vnic...\n")
            try:
                sock = socket.create_connection((args[0], 443))
                if sock:
                    sock.close
            except:
                pass

            self.sessionid = options.sessionid
            self.login_otp = options.login_otp

            if "blobstore" in self.url and (options.waitforOTP or options.login_otp):
                options.waitforOTP = None
                options.login_otp = None
                self.rdmc.ui.printer(
                    "Warning: For local inband mode, TFA is not supported, options --wait_for_otp "
                    "and --otp will be ignored\n"
                )

            if options.waitforOTP:
                try:
                    self.perform_login(options, skipbuild, user_ca_cert_data)
                except redfish.rest.connections.OneTimePasscodeError:
                    self.rdmc.ui.printer("One Time Passcode Sent to registered email.\n")
                    ans = input("Enter OTP: ")
                    self.login_otp = ans
                    self.perform_login(options, skipbuild, user_ca_cert_data)
            else:
                self.perform_login(options, skipbuild, user_ca_cert_data)
        except ServerDownOrUnreachableError as excp:
            self.rdmc.ui.printer("The following error occurred during login: '%s'\n" % str(excp.__class__.__name__))

        self.username = None
        self.password = None

        # Warning for cache enabled, since we save session in plain text
        if not self.rdmc.encoding:
            self.rdmc.ui.warn("Cache is activated. Session keys are stored in plaintext.")

        if self.rdmc.opts.debug:
            self.rdmc.ui.warn("Logger is activated. Logging is stored in plaintext.")

        if options.selector:
            try:
                self.rdmc.app.select(selector=options.selector)

                if self.rdmc.opts.verbose:
                    self.rdmc.ui.printer(("Selected option: '%s'\n" % options.selector))
            except Exception as excp:
                raise redfish.ris.InstanceNotFoundError(excp)

    def loginvalidation(self, options, args):
        """Login helper function for login validations

        :param options: command line options
        :type options: list.
        :param args: command line arguments
        :type args: list.
        """
        # Fill user name/password from config file
        if not options.user:
            options.user = self.rdmc.config.username
        if not options.password:
            options.password = self.rdmc.config.password
        if not hasattr(options, "user_certificate"):
            options.user_certificate = self.rdmc.config.user_cert
        if not hasattr(options, "user_root_ca_key"):
            options.user_root_ca_key = self.rdmc.config.user_root_ca_key
        if not hasattr(options, "user_root_ca_password"):
            options.user_root_ca_password = self.rdmc.config.user_root_ca_password

        if (
            options.user
            and not options.password
            and (
                not hasattr(options, "user_certificate")
                or not hasattr(options, "user_root_ca_key")
                or hasattr(options, "user_root_ca_password")
            )
        ):
            # Option for interactive entry of password
            tempinput = getpass.getpass().rstrip()
            if tempinput:
                options.password = tempinput
            else:
                raise InvalidCommandLineError("Empty or invalid password was entered.")

        if options.user:
            self.username = options.user

        if options.password:
            self.password = options.password

        if options.encode:
            self.username = Encryption.decode_credentials(self.username).decode("utf-8")
            self.password = Encryption.decode_credentials(self.password).decode("utf-8")

        if options.biospassword:
            self.biospassword = options.biospassword

        # Assignment of url in case no url is entered
        if getattr(options, "force_vnic", False):
            if not (getattr(options, "ca_cert_bundle", False) or getattr(options, "user_certificate", False)):
                if not (self.username and self.password) and not options.sessionid:
                    raise UsernamePasswordRequiredError("Please provide credentials to login with VNIC")
            self.url = "https://16.1.15.1"
        else:
            self.url = "blobstore://."

        if args:
            # Any argument should be treated as an URL
            self.url = args[0]

            # Verify that URL is properly formatted for https://
            if "https://" not in self.url:
                self.url = "https://" + self.url

            if not (
                hasattr(options, "user_certificate")
                or hasattr(options, "user_root_ca_key")
                or hasattr(options, "user_root_ca_password")
            ):
                if not (options.username and options.password):
                    raise InvalidCommandLineError("Empty username or password was entered.")
        else:
            # Check to see if there is a URL in config file
            if self.rdmc.config.url:
                self.url = self.rdmc.config.url

    def definearguments(self, customparser):
        """Wrapper function for new command main function

        :param customparser: command line input
        :type customparser: parser.
        """

        def remove_argument(parser, arg):
            for action in parser._actions:
                opts = action.option_strings
                if (opts and opts[0] == arg) or action.dest == arg:
                    parser._remove_action(action)
                    break

            for action in parser._action_groups:
                for group_action in action._group_actions:
                    opts = group_action.option_strings
                    if (opts and opts[0] == arg) or group_action.dest == arg:
                        action._group_actions.remove(group_action)
                        return

        if not customparser:
            return

        customparser.add_argument(
            "--wait_for_otp",
            dest="waitforOTP",
            help="Optionally include this flag to implicitly wait for OTP.",
            action="store_true",
            default=None,
        )
        self.cmdbase.add_login_arguments_group(customparser)
        remove_argument(customparser, "url")
        customparser.add_argument(
            "--selector",
            dest="selector",
            help="Optionally include this flag to select a type to run"
            " the current command on. Use this flag when you wish to"
            " select a type without entering another command, or if you"
            " wish to work with a type that is different from the one"
            " you currently have selected.",
            default=None,
        )



###
# Copyright 2019 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
""" Certificates Command for rdmc """

import copy
import json
import re
from argparse import RawDescriptionHelpFormatter
from collections import OrderedDict

import redfish.ris
from redfish.ris.ris import RisInstanceNotFoundError
from redfish.ris.rmc_helper import (
    IloResponseError,
    IncompatibleiLOVersionError,
    InvalidPathError,
)
from redfish.ris.utils import diffdict, json_traversal, json_traversal_delete_empty

try:
    from rdmc_helper import (
        UI,
        Encryption,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        InvalidFileInputError,
        InvalidPropertyError,
        NoDifferencesFoundError,
        RdmcError,
        ReturnCodes,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        UI,
        Encryption,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        InvalidFileInputError,
        InvalidPropertyError,
        NoDifferencesFoundError,
        RdmcError,
        ReturnCodes,
    )

__eth_file__ = "eth.json"
__subparsers__ = ["save", "load"]


class EthernetCommand:
    """ Commands for Ethernet Management Controller and RDE supported NIC card \
        configuration on server """

    def __init__(self):
        self.ident = {
            "name": "ethernet",
            "usage": None,
            "description": "Save or load from a JSON formatted file containing "
            "properties pertaining to a system's iLO ethernet management controller. "
            "Additionally enable/disable individual management controller NICs (including VNIC), "
            "IPv4 and IPv6 addressing configuration and domain name servers."
            "\n\tBy default the JSON file will be named 'eth.json'.\n\t"
            "\n\t***Credentials for an iLO administrator level account must be provided as"
            " well as the iLO base URL.***"
            "\n\n\tSave ethernet management controller data.\n\texample: ethernet save "
            "\n\n\tLoad ethernet management controller data.\n\texample: ethernet load"
            "\n\n\tSave ethernet management controller data to a different file, silently\n\t"
            "ethernet save -f <filename> --silent"
            "\n\n\tLoad etherent management controller data from a different file, silently\n\t"
            "ethernet load -f <filename> --silent"
            "\n\n\tEnable network interfaces by listing each interface to be enabled. **Note**: "
            "Non-existent interfaces will be omitted from configuration.\n\t "
            "ethernet --enable_nic 1,2,3"
            "\n\n\tDisable network interfaces by listing each interface to be disabled. **Note**: "
            "Non-existent interfaces will be omitted from configuration.\n\t "
            "ethernet --disable_nic 1,2,3"
            "\n\n\tEnable virtual network interface of management network.\n\t"
            "ethernet --enable_vnic"
            "\n\n\tDisable virtual network interface of management network.\n\t  "
            "ethernet --disable_vnic"
            "\n\n\tEnable Enhanced Download Performance.\n\t"
            "ethernet --enable_enhanced_downloads"
            "\n\n\tDisable Enhanced Download Performance.\n\t  "
            "ethernet --disable_enhanced_downloads"
            "\n\n\tConfigure Domain Name Servers (DNS) in a list: <DNS1> <DNS2> <DNS3>\n\t"
            "ethernet --nameservers 8.8.8.8,1.1.1.1,2.2.2.2 OR ethernet --nameservers "
            "dns_resolver1.aws.com,dns_resolver2.aws.com"
            "\n\n\tConfigure Static IPv4 Settings. Provide a list of network settings\n\t"
            "ethernet --network_ipv4 <ipv4 address>,<ipv4 gateway>,<ipv4 network mask>"
            "\n\n\tConfigure Proxy Settings. Provide a proxy server and port\n\t"
            "ethernet --proxy http://proxy.company.net:8080"
            "\n\n\tClear Proxy Settings. \n\t"
            "ethernet --proxy None"
            "\n\n\tConfigure Static IPv6 Settings. Provide a list of network settings\n\t"
            "ethernet --network_ipv6 <ipv6 address>,<ipv6 gateway>,<ipv6 network mask>",
            "summary": "Command for configuring Ethernet Management Controller Interfaces and " "associated properties",
            "aliases": [],
            "auxcommands": ["IloResetCommand"],
        }
        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()
        self.save = None
        self.load = None
        self.eth_file = None  # set in validation
        self.eth_args = None
        self.saved_inputline = None  # set in validation

    def run(self, line, help_disp=False):
        """Main Ethernet Management Controller Interfaces Run
        :param line: string of arguments passed in
        :type line: str.
        :param help_disp: display help flag
        :type line: bool.
        """
        if help_disp:
            line.append("-h")
            try:
                (_, _) = self.rdmc.rdmc_parse_arglist(self, line)
            except:
                return ReturnCodes.SUCCESS
            return ReturnCodes.SUCCESS
        try:
            ident_subparser = False
            for cmnd in __subparsers__:
                if cmnd in line:
                    (options, args) = self.rdmc.rdmc_parse_arglist(self, line)
                    ident_subparser = True
                    break
            if not ident_subparser:
                (options, args) = self.rdmc.rdmc_parse_arglist(self, line, default=True)
        except:
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")

        self.ethernetvalidation(options)

        if "default" in options.command.lower():
            # flags are used to showing data with respect to enhance download
            flag = True
            return_flag = True
            if options.enable_enhanced_downloads or options.disable_enhanced_downloads:
                ilo_ver = self.rdmc.app.getiloversion()
                self.rdmc.ui.printer("iLO Version is " + str(ilo_ver) + "\n")
                if ilo_ver < 5.263:
                    raise IncompatibleiLOVersionError("Enhance download required minimum " "iLO Version iLO5.263")
                # net_path = self.rdmc.app.getidbytype('NetworkProtocol.')
                net_path = self.rdmc.app.typepath.defs.managerpath + "NetworkProtocol"
                results = self.rdmc.app.get_handler(net_path, silent=True, service=True).dict
                body = dict()
                body["Oem"] = {}
                body["Oem"]["Hpe"] = {}
                if options.enable_enhanced_downloads:
                    if results["Oem"]["Hpe"]["EnhancedDownloadPerformanceEnabled"]:
                        self.rdmc.ui.printer(
                            "Enhanced Download Performance " "already enabled!! \n",
                            verbose_override=True,
                        )
                        return_flag = False
                    else:
                        self.rdmc.ui.printer("Enabling enhanced " "downloads...\n", verbose_override=True)
                        body["Oem"]["Hpe"]["EnhancedDownloadPerformanceEnabled"] = True
                        flag = False
                else:
                    if not results["Oem"]["Hpe"]["EnhancedDownloadPerformanceEnabled"]:
                        self.rdmc.ui.printer(
                            "Enhanced Download Performance " "already disabled!!\n\n",
                            verbose_override=True,
                        )
                        return_flag = False
                    else:
                        self.rdmc.ui.printer("Disabling enhanced " "downloads...\n", verbose_override=True)
                        body["Oem"]["Hpe"]["EnhancedDownloadPerformanceEnabled"] = False
                        flag = False
                if options.enable_vnic or options.disable_vnic:
                    self.rdmc.app.patch_handler(net_path, body, service=False, silent=True)
                elif return_flag:
                    self.rdmc.app.patch_handler(net_path, body, service=False, silent=False)
            if options.enable_vnic or options.disable_vnic:
                results = self.rdmc.app.get_handler(
                    self.rdmc.app.typepath.defs.managerpath, service=True, silent=True
                ).dict
                body = dict()
                body["Oem"] = {}
                body["Oem"]["Hpe"] = {}
                if options.enable_vnic:
                    if results["Oem"]["Hpe"]["VirtualNICEnabled"]:
                        self.rdmc.ui.printer("Virtual NIC already enabled!!\n", verbose_override=True)
                        return ReturnCodes.SUCCESS
                    self.rdmc.ui.printer("Enabling Virtual NIC...\n", verbose_override=True)
                    body["Oem"]["Hpe"]["VirtualNICEnabled"] = True
                else:
                    if not results["Oem"]["Hpe"]["VirtualNICEnabled"]:
                        self.rdmc.ui.printer("Virtual NIC already disabled!!\n", verbose_override=True)
                        return ReturnCodes.SUCCESS
                    self.rdmc.ui.printer("Disabling Virtual NIC...\n", verbose_override=True)
                    body["Oem"]["Hpe"]["VirtualNICEnabled"] = False
                self.rdmc.app.patch_handler(
                    self.rdmc.app.typepath.defs.managerpath,
                    body,
                    service=False,
                    silent=False,
                )
                self.rdmc.ui.printer("Warning: Resetting iLO...\n")
                self.auxcommands["iloreset"].run("")
                self.rdmc.ui.printer("You will need to re-login to access this system...\n")
            elif flag and return_flag:
                data = self.get_data(options)
                get_data = True
                for inst in data:
                    if "ethernetinterface" in inst.lower() or "ethernetnetworkinterface" in inst.lower():
                        for path in data[inst]:
                            if (
                                "managers/1/ethernetinterfaces/1" in path.lower()
                                or "managers/1/ethernetnetworkinterfaces/1" in path.lower()
                            ):  # only process managers interfaces
                                if options.enable_nic or options.disable_nic:
                                    get_data = False
                                    self.enable_disable_nics(data[inst][path], options)

                                if (
                                    options.network_ipv4 or options.network_ipv6 or options.nameservers or options.proxy
                                ) and "virtual" not in data[inst][path]["Name"].lower():
                                    get_data = False
                                    self.configure_static_settings(data[inst][path], options)
                                    if options.proxy:
                                        return ReturnCodes.SUCCESS
                    if "networkprotocol" in inst.lower():
                        for path in data[inst]:
                            if "managers/1/networkprotocol" in path.lower():
                                if options.enable_enhanced_downloads or options.disable_enhanced_downloads:
                                    get_data = False
                if get_data:
                    self.output_data(data, options, get_data)
                else:
                    self.load_main_function(options, data)

        if "save" in options.command.lower():
            self.save = True
            self.load = False
            if "save" in options.command.lower() and not options.ethfilename:
                self.rdmc.ui.printer("Saving configurations to default file '%s'.\n" % self.eth_file)
            self.output_data(self.get_data(options), options, False)

        elif "load" in options.command.lower():
            self.save = False
            self.load = True
            if options.force_network_config:
                self.load_main_function(options)
            else:
                self.rdmc.ui.printer("Skipping network configurations as " "--force_network_config is not included.\n")

        return ReturnCodes.SUCCESS

    def get_data(self, options):
        """Obtain data for Ethernet Management Interfaces, Ethernet System Interfaces, \
           Manager Network Services and iLO Date Time Types and subsequently save to a dictionary \
        :param options: command line options
        :type options: attribute
        """

        network_data_collection = eth_iface = manager_network_services = ilo_date_time = OrderedDict()

        if self.rdmc.app.redfishinst.is_redfish:
            try:
                iface_results = self.rdmc.app.select(selector="ethernetinterface.")
            except RisInstanceNotFoundError:
                self.rdmc.ui.printer("Type 'EthernetInterface.' not available.\n")
        else:
            try:
                iface_results = self.rdmc.app.select(selector="ethernetnetworkinterface.")
            except RisInstanceNotFoundError:
                self.rdmc.ui.printer("Type 'EthernetNetworkInterface.' not available.\n")

        for inst in iface_results:
            try:
                eth_iface[inst.type].update({inst.path: inst.dict})
            except KeyError:
                eth_iface[inst.type] = {inst.path: inst.dict}
                continue
            except AttributeError:
                self.rdmc.ui.printer("Missing instance data for type '%s' : '%s'" % (inst.type, inst.path))
        if self.rdmc.app.typepath.defs.isgen9:
            selector_content = self.rdmc.app.select(selector="ManagerNetworkService.0.11.1.networkprotocol")
        else:
            selector_content = self.rdmc.app.select(selector="networkprotocol.")
        for inst in selector_content:
            try:
                manager_network_services[inst.type].update({inst.path: inst.dict})
            except KeyError:
                manager_network_services[inst.type] = {inst.path: inst.dict}
                continue
            except AttributeError:
                self.rdmc.ui.printer("Missing instance data for type '%s' : '%s'" % (inst.type, inst.path))

        for inst in self.rdmc.app.select(selector="datetime."):
            try:
                ilo_date_time[inst.type].update({inst.path: inst.dict})
            except KeyError:
                ilo_date_time[inst.type] = {inst.path: inst.dict}
                continue
            except AttributeError:
                self.rdmc.ui.printer("Missing instance data for type '%s' : '%s'" % (inst.type, inst.path))

        network_data_collection.update(eth_iface)
        network_data_collection.update(manager_network_services)
        network_data_collection.update(ilo_date_time)

        return network_data_collection

    def output_data(self, data, options, get_only=False):
        """ Output data in requested json file or on console
        :param data: dictionary containing ethernet configuration data:
        :type data: dictionary
        :param options: command line options
        :type options: attribute
        :param get_only: parameter to only output to console (do not save data to file). \
        Essentially a targeted \'get\' command.
        :type: boolean
        """

        outdata = list()

        outdata.append(self.rdmc.app.create_save_header())

        for _type in data:
            for path in data.get(_type):
                temp = dict()
                try:
                    temp[_type.split("#")[-1]].update({path: self.rdmc.app.removereadonlyprops(data[_type][path])})
                    outdata.append(temp)
                except KeyError:
                    temp[_type.split("#")[-1]] = {path: self.rdmc.app.removereadonlyprops(data[_type][path])}
                    outdata.append(temp)
                except Exception:
                    pass

        if self.eth_file and not get_only:
            if options.encryption:
                with open(self.eth_file, "wb") as outfile:
                    outfile.write(
                        Encryption().encrypt_file(
                            json.dumps(outdata, indent=2, cls=redfish.ris.JSONEncoder),
                            options.encryption,
                        )
                    )
            else:
                with open(self.eth_file, "w") as outfile:
                    outfile.write(json.dumps(outdata, indent=2, cls=redfish.ris.JSONEncoder))
        else:
            if options.json:
                self.rdmc.ui.print_out_json_ordered(outdata)
            else:
                UI().print_out_human_readable(outdata)

    def enable_disable_nics(self, data, options):
        if options.disable_nic:
            for ident in re.split("[, ]", options.disable_nic):
                try:
                    if int(data.get("Id")) == int(ident):
                        data.update({"InterfaceEnabled": False})
                        break
                except ValueError:
                    if data.get("Name") == ident:
                        data.update({"InterfaceEnabled": False})
                        break

        if options.enable_nic:
            for ident in re.split("[, ]", options.enable_nic):
                try:
                    if int(data.get("Id")) == int(ident):
                        data.update({"InterfaceEnabled": True})
                        break
                except ValueError:
                    if data.get("Name") == ident:
                        data.update({"InterfaceEnabled": True})
                        break

    def configure_static_settings(self, data, options):
        if options.network_ipv4:
            usr_data = re.split("[, ]", options.network_ipv4)
            if len(usr_data) > 2:
                data["IPv4Addresses"][0] = {
                    "Address": usr_data[0],
                    "Gateway": usr_data[1],
                    "SubnetMask": usr_data[2],
                }
                data["DHCPv4"].update({"DHCPEnabled": False})
                data["Oem"][self.rdmc.app.typepath.defs.oemhp]["DHCPv4"].update({"Enabled": False})
            else:
                raise InvalidCommandLineErrorOPTS(
                    "An invalid number of arguments provided to "
                    " quick networking configuration. Check '--network_ipv4' entry."
                )
        else:
            del data["IPv4Addresses"]

        if options.network_ipv6:
            usr_data = re.split("[, ]", options.network_ipv6)
            if len(usr_data) > 2:
                next(iter(data["IPv6Addresses"])).update({"Address": usr_data[0], "PrefixLength": usr_data[2]})
                data["IPv6DefaultGateway"] = options.network_ipv6[1]
                data["DHCPv6"].update({"DHCPEnabled": False})
                data["Oem"][self.rdmc.app.typepath.defs.oemhp]["DHCPv6"].update({"Enabled": False})
            else:
                raise InvalidCommandLineErrorOPTS(
                    "An invalid number of arguments provided to "
                    " quick networking configuration. Check '--network_ipv6' entry."
                )
        else:
            del data["IPv6Addresses"]

        if options.nameservers:
            usr_data = re.split("[, ]", options.nameservers)
            if len(usr_data) <= 3:
                ipv6_list = list()
                ipv4_list = list()
                static_list = list()
                for elem in usr_data:
                    oem_ipv4 = data["Oem"][self.rdmc.app.typepath.defs.oemhp]["IPv4"]
                    oem_ipv6 = data["Oem"][self.rdmc.app.typepath.defs.oemhp]["IPv6"]
                    if "::" in elem:
                        ipv6_list.append(elem)
                    elif "." in elem:
                        ipv4_list.append(elem)
                    static_list.append(elem)
                oem_ipv6["DNSServers"] = ipv6_list
                oem_ipv4["DNSServers"] = ipv4_list
                data["StaticNameServers"] = static_list
            else:
                raise InvalidCommandLineErrorOPTS("Name Servers argument has to be less than or equal to 2")
        else:
            del data["StaticNameServers"]
            del data["Oem"][self.rdmc.app.typepath.defs.oemhp]["IPv4"]["DNSServers"]
            del data["Oem"][self.rdmc.app.typepath.defs.oemhp]["IPv6"]["DNSServers"]

        if options.proxy and options.proxy != "None":
            body = dict()
            body["Oem"] = {}
            body["Oem"]["Hpe"] = {}
            body["Oem"]["Hpe"]["WebProxyConfiguration"] = {}
            proxy_body = body["Oem"]["Hpe"]["WebProxyConfiguration"]
            proxy_body["ProxyServer"] = None
            proxy_body["ProxyUserName"] = None
            proxy_body["ProxyPassword"] = None
            if "https" in options.proxy:
                proxy_body["ProxyPort"] = 443
            else:
                proxy_body["ProxyPort"] = 80
            if "@" in options.proxy:
                proxy = options.proxy.split("@")
                proxy_usr_pass = proxy[0]
                proxy_srv_port = proxy[1]
                if "//" in proxy_usr_pass:
                    proxy_usr_pass = proxy_usr_pass.split("//")[1]
                if ":" in proxy_srv_port:
                    proxy = proxy_srv_port.split(":")
                    proxy_body["ProxyServer"] = proxy[0]
                    proxy_body["ProxyPort"] = int(proxy[1])
                else:
                    proxy_body["ProxyServer"] = proxy_srv_port
                if ":" in proxy_usr_pass:
                    proxy = proxy_usr_pass.split(":")
                    proxy_body["ProxyPassword"] = proxy[1]
                    proxy_body["ProxyUserName"] = proxy[0]
                else:
                    proxy_body["ProxyUserName"] = proxy_usr_pass
            else:
                proxy_srv_port = options.proxy
                if "//" in proxy_srv_port:
                    proxy_srv_port = proxy_srv_port.split("//")[1]
                if ":" in proxy_srv_port:
                    proxy = proxy_srv_port.split(":")
                    proxy_body["ProxyServer"] = proxy[0]
                    proxy_body["ProxyPort"] = int(proxy[1])
                else:
                    proxy_body["ProxyServer"] = proxy_srv_port
            path = self.rdmc.app.getidbytype("NetworkProtocol.")

            if path and body:
                self.rdmc.ui.printer("Enabling Proxy configuration...\n", verbose_override=True)
                self.rdmc.app.patch_handler(path[0], body, service=False, silent=False)
        elif options.proxy and options.proxy == "None":
            body = dict()
            body["Oem"] = {}
            body["Oem"]["Hpe"] = {}
            body["Oem"]["Hpe"]["WebProxyConfiguration"] = {}
            proxy_body = body["Oem"]["Hpe"]["WebProxyConfiguration"]
            proxy_body["ProxyServer"] = ""
            proxy_body["ProxyPort"] = None
            proxy_body["ProxyUserName"] = ""
            proxy_body["ProxyPassword"] = None
            path = self.rdmc.app.getidbytype("NetworkProtocol.")

            if path and body:
                self.rdmc.ui.printer("Clearing Proxy configuration...\n", verbose_override=True)
                self.rdmc.app.patch_handler(path[0], body, service=False, silent=False)

        json_traversal_delete_empty(data, None, None)

    def load_main_function(self, options, data=None):
        """
        Load main function. Handles SSO and SSL/TLS Certificates, kickoff load
        helper, load of patch file, get server status and issues system reboot
        and iLO reset after completion of all post and patch commands.
        :param options: command line options
        :type options: attribute
        :param data: data
        :type data: optional
        """

        if self.rdmc.app.redfishinst.is_redfish:
            _ = "ethernetinterface."
        else:
            _ = "ethernetnetworkinterface."

        if not data:
            data = {}
            try:
                if options.encryption:
                    with open(self.eth_file, "rb") as file_handle:
                        data = json.loads(Encryption().decrypt_file(file_handle.read(), options.encryption))
                else:
                    with open(self.eth_file, "rb") as file_handle:
                        data = json.loads(file_handle.read())
            except:
                raise InvalidFileInputError(
                    "Invalid file formatting found. Verify the file has a " "valid JSON format."
                )
        if "load" in options.command.lower():
            for d in data:
                for ilotype, subsect in d.items():
                    _type = ilotype.split(".")[0]
                    for _path in subsect:
                        if not subsect[_path]:
                            continue
                        elif "ethernetinterface" in _type.lower() or "ethernetnetworkinterface" in _type.lower():
                            if "managers" in _path.lower():
                                self.load_ethernet_aux(_type, _path, d[ilotype][_path])
                            elif "systems" in _path.lower():
                                self.rdmc.ui.warn("Systems Ethernet Interfaces '%s' " "cannot be modified." % _path)
                                continue
                        elif "datetime" in _type.lower():
                            if "StaticNTPServers" in list(subsect.get(_path).keys()):
                                # must set NTP Servers to static in OEM then reset iLO for StaticNTPServers
                                # property to appear in iLODateTime
                                if self.rdmc.app.redfishinst.is_redfish:
                                    eth_config_type = "ethernetinterface"
                                else:
                                    eth_config_type = "ethernetnetworkinterface"
                                for key in list(data.keys()):
                                    if key.split(".")[0].lower() == eth_config_type:
                                        eth_config_type = key
                                        for _path in data[eth_config_type]:
                                            if "managers" in _path.lower():
                                                try:
                                                    oemhp = self.rdmc.app.typepath.defs.oemhp
                                                    data[eth_config_type][_path]["DHCPv4"]["UseNTPServers"] = True
                                                    data[eth_config_type][_path]["DHCPv6"]["UseNTPServers"] = True
                                                    data[eth_config_type][_path]["Oem"][oemhp]["DHCPv4"][
                                                        "UseNTPServers"
                                                    ] = True
                                                    data[eth_config_type][_path]["Oem"][oemhp]["DHCPv6"][
                                                        "UseNTPServers"
                                                    ] = True
                                                    self.load_ethernet_aux(
                                                        eth_config_type,
                                                        _path,
                                                        data[eth_config_type][_path],
                                                    )
                                                except KeyError:
                                                    self.rdmc.ui.printer(
                                                        "Unable to configure " "'UseNTPServers' for '%s'.\n" % _path
                                                    )
                                                self.rdmc.ui.printer(
                                                    "iLO must be reset in order for "
                                                    "changes to static network time protocol servers to "
                                                    "take effect.\n"
                                                )
        elif "default" in options.command.lower():
            for ilotype, subsect in data.items():
                _type = ilotype.split(".")[0]
                for _path in subsect:
                    if not subsect[_path]:
                        continue
                    elif "ethernetinterface" in _type.lower() or "ethernetnetworkinterface" in _type.lower():
                        if "managers" in _path.lower():
                            self.load_ethernet_aux(_type, _path, data[ilotype][_path])
                        elif "systems" in _path.lower():
                            self.rdmc.ui.warn("Systems Ethernet Interfaces '%s' " "cannot be modified." % _path)
                            continue
                    elif "datetime" in _type.lower():
                        if "StaticNTPServers" in list(subsect.get(_path).keys()):
                            # must set NTP Servers to static in OEM then reset iLO for StaticNTPServers
                            # property to appear in iLODateTime
                            if self.rdmc.app.redfishinst.is_redfish:
                                eth_config_type = "ethernetinterface"
                            else:
                                eth_config_type = "ethernetnetworkinterface"
                            for key in list(data.keys()):
                                if key.split(".")[0].lower() == eth_config_type:
                                    eth_config_type = key
                                    for _path in data[eth_config_type]:
                                        if "managers" in _path.lower():
                                            try:
                                                oemhp = self.rdmc.app.typepath.defs.oemhp
                                                data[eth_config_type][_path]["DHCPv4"]["UseNTPServers"] = True
                                                data[eth_config_type][_path]["DHCPv6"]["UseNTPServers"] = True
                                                data[eth_config_type][_path]["Oem"][oemhp]["DHCPv4"][
                                                    "UseNTPServers"
                                                ] = True
                                                data[eth_config_type][_path]["Oem"][oemhp]["DHCPv6"][
                                                    "UseNTPServers"
                                                ] = True
                                                self.load_ethernet_aux(
                                                    eth_config_type,
                                                    _path,
                                                    data[eth_config_type][_path],
                                                )
                                            except KeyError:
                                                self.rdmc.ui.printer(
                                                    "Unable to configure " "'UseNTPServers' for '%s'.\n" % _path
                                                )
                                            self.rdmc.ui.printer(
                                                "iLO must be reset in order for "
                                                "changes to static network time protocol servers to "
                                                "take effect.\n"
                                            )

    def load_ethernet_aux(self, _type, _path, ethernet_data):
        """helper function for parsing and bucketting ethernet properties.
        :param _type: Redfish type used for querying the current server data
        :type _type: string
        :param _path: URI path to be patched.
        :type _path: string
        :param ethernet_data: JSON containing the ethernet instance (and valid, associated
                         properties) to be patched.
        :type ethernet_data: JSON

        """

        support_ipv6 = True
        dhcpv4curr = dhcpv4conf = oem_dhcpv4conf = dict()
        _ = dhcpv6conf = oem_dhcpv6conf = dict()
        errors = []

        ident_eth = False
        if "EthernetInterface" in _type:
            for curr_sel in self.rdmc.app.select(
                _type.split(".")[0] + ".",
                (
                    self.rdmc.app.typepath.defs.hrefstring,
                    self.rdmc.app.typepath.defs.managerpath + "*",
                ),
                path_refresh=True,
            ):
                if curr_sel.path == _path:
                    ident_eth = True
                    break
            # 'links/self/href' required when using iLO 4 (rest).
        elif "EthernetNetworkInterface" in _type:
            for curr_sel in self.rdmc.app.select(
                _type.split(".")[0] + ".",
                (
                    "links/self/" + self.rdmc.app.typepath.defs.hrefstring,
                    self.rdmc.app.typepath.defs.managerpath + "*",
                ),
                path_refresh=True,
            ):
                if curr_sel.path == _path:
                    ident_eth = True
                    break
        else:
            raise Exception("Invalid type in management NIC load operation: '%s'" % _type)

        if not ident_eth:
            raise InvalidPathError("Path: '%s' is invalid/not identified on this server.\n" % _path)

        ident_name = curr_sel.dict.get("Name")
        ident_id = curr_sel.dict.get("Id")
        # ENABLING ETHERNET INTERFACE SECTION
        try:
            # Enable the Interface if called for and not already enabled
            if ethernet_data.get("InterfaceEnabled") and not curr_sel.dict.get("InterfaceEnabled"):
                self.rdmc.app.patch_handler(_path, {"InterfaceEnabled": True}, silent=True)
                self.rdmc.ui.printer("NIC Interface Enabled.\n")
            # Disable the Interface if called for and not disabled already
            # No need to do anything else, just return
            elif not ethernet_data.get("InterfaceEnabled") and not curr_sel.dict.get("InterfaceEnabled"):
                self.rdmc.app.patch_handler(_path, {"InterfaceEnabled": False}, silent=True)
                self.rdmc.ui.warn("NIC Interface Disabled. All additional configurations " "omitted.")
                return
        except (KeyError, NameError, TypeError, AttributeError):
            # check OEM for NICEnabled instead
            if (
                not curr_sel.dict["Oem"][self.rdmc.app.typepath.defs.oemhp]["NICEnabled"]
                and ethernet_data["Oem"][self.rdmc.app.typepath.defs.oemhp]["NICEnabled"]
            ):
                self.rdmc.app.patch_handler(
                    _path,
                    {"Oem": {self.rdmc.app.typepath.defs.oemhp: {"NICEnabled": True}}},
                    silent=True,
                )
                self.rdmc.ui.printer("NIC Interface Enabled.\n")
            elif (
                not curr_sel.dict["Oem"][self.rdmc.app.typepath.defs.oemhp]["NICEnabled"]
                and not ethernet_data["Oem"][self.rdmc.app.typepath.defs.oemhp]["NICEnabled"]
            ):
                self.rdmc.app.patch_handler(
                    _path,
                    {"Oem": {self.rdmc.app.typepath.defs.oemhp: {"NICEnabled": False}}},
                    silent=True,
                )
                self.rdmc.ui.printer("NIC Interface Disabled.\n")
                return
        # except IloResponseError should just be raised and captured by decorator. No point in
        # performing any other operations if the interface can not be set.

        # END ENABLING ETHERNET INTEFACE SECTION
        # ---------------------------------------
        # DETERMINE DHCPv4 and DHCPv6 States and associated flags

        if "NICSupportsIPv6" in list(curr_sel.dict["Oem"][self.rdmc.app.typepath.defs.oemhp].keys()):
            support_ipv6 = curr_sel.dict["Oem"][self.rdmc.app.typepath.defs.oemhp]["NICSupportsIPv6"]

        # obtain DHCPv4 Config and OEM
        try:
            if "DHCPv4" in list(curr_sel.dict.keys()) and "DHCPv4" in list(ethernet_data.keys()):
                dhcpv4curr = copy.deepcopy(curr_sel.dict["DHCPv4"])
                dhcpv4conf = copy.deepcopy(ethernet_data["DHCPv4"])
        except (KeyError, NameError, TypeError, AttributeError):
            errors.append("Unable to find Redfish DHCPv4 Settings.\n")
        finally:
            try:
                oem_dhcpv4conf = copy.deepcopy(ethernet_data["Oem"][self.rdmc.app.typepath.defs.oemhp]["DHCPv4"])

            except (KeyError, NameError):
                errors.append("Unable to find OEM Keys for DHCPv4 or IPv4")

        try:
            if support_ipv6:
                if "DHCPv6" in list(curr_sel.dict.keys()) and "DHCPv6" in list(ethernet_data.keys()):
                    dhcpv6conf = copy.deepcopy(ethernet_data["DHCPv6"])
            else:
                self.rdmc.ui.warn("NIC Does not support IPv6.")
        except (KeyError, NameError, TypeError, AttributeError):
            errors.append("Unable to find Redfish DHCPv6 Settings.\n")
        finally:
            try:
                oem_dhcpv6conf = copy.deepcopy(ethernet_data["Oem"][self.rdmc.app.typepath.defs.oemhp]["DHCPv6"])
            except (KeyError, NameError):
                errors.append("Unable to find OEM Keys for DHCPv6 or IPv6")

        try:
            # if DHCP Enable request but not currently enabled
            if dhcpv4conf.get("DHCPEnabled") and not curr_sel.dict["DHCPv4"]["DHCPEnabled"]:
                self.rdmc.app.patch_handler(_path, {"DHCPv4": {"DHCPEnabled": True}}, silent=True)
                self.rdmc.ui.printer("DHCP Enabled.\n")
            # if DHCP Disable request but currently enabled
            elif not dhcpv4conf["DHCPEnabled"] and curr_sel.dict["DHCPv4"]["DHCPEnabled"]:
                self.rdmc.app.patch_handler(_path, {"DHCPv4": {"DHCPEnabled": False}}, silent=True)
                dhcpv4conf["UseDNSServers"] = False
                dhcpv4conf["UseNTPServers"] = False
                dhcpv4conf["UseGateway"] = False
                dhcpv4conf["UseDomainName"] = False
                self.rdmc.ui.printer("DHCP Disabled.\n")
        except (KeyError, NameError, TypeError, AttributeError):
            # try with OEM
            try:
                if (
                    oem_dhcpv4conf.get("Enabled")
                    and not curr_sel.dict["Oem"][self.rdmc.app.typepath.defs.oemhp]["DHCPv4"]["Enabled"]
                ):
                    self.rdmc.app.patch_handler(
                        _path,
                        {"Oem": {self.rdmc.app.typepath.defs.oemhp: {"DHCPv4": {"DHCPEnabled": True}}}},
                        silent=True,
                    )
                    self.rdmc.ui.printer("DHCP Enabled.\n")
                    if "IPv4Addresses" in ethernet_data:
                        del ethernet_data["IPv4Addresses"]
                elif (
                    not oem_dhcpv4conf.get("Enabled")
                    and curr_sel.dict["Oem"][self.rdmc.app.typepath.defs.oemhp]["DHCPv4"]["Enabled"]
                ):
                    oem_dhcpv4conf["UseDNSServers"] = False
                    oem_dhcpv4conf["UseNTPServers"] = False
                    oem_dhcpv4conf["UseGateway"] = False
                    oem_dhcpv4conf["UseDomainName"] = False
                    self.rdmc.ui.printer("DHCP Disabled.\n")
            except (KeyError, NameError) as exp:
                errors.append("Failure in parsing or removing data in OEM DHCPv4: %s.\n" % exp)

        try:
            # if the ClientIDType is custom and we are missing the ClientID then this property can
            # not be set.
            if "ClientIdType" in list(dhcpv4conf.keys()):
                if dhcpv4conf["ClientIdType"] == "Custom" and "ClientID" not in list(dhcpv4conf.keys()):
                    del ethernet_data["DHCPv4"]["ClientIdType"]
            elif "ClientIdType" in list(oem_dhcpv4conf.keys()):
                if oem_dhcpv4conf["ClientIdType"] == "Custom" and "ClientID" not in list(oem_dhcpv4conf.keys()):
                    del ethernet_data["Oem"][self.rdmc.app.typepath.defs.oemhp]["DHCPv4"]["ClientIdType"]
        except (KeyError, NameError, TypeError, AttributeError):
            try:
                if "ClientIdType" in list(oem_dhcpv4conf.keys()):
                    if oem_dhcpv4conf["ClientIdType"] == "Custom" and "ClientID" not in list(oem_dhcpv4conf.keys()):
                        del ethernet_data["Oem"][self.rdmc.app.typepath.defs.oemhp]["DHCPv4"]["ClientIdType"]
            except (KeyError, NameError) as exp:
                errors.append("Unable to remove property %s.\n" % exp)

        # special considerations go here for things that need to stay despite diffdict
        # EX: IPv4 addresses (aka bug). Changing only one property within the
        # IPv4StaticAddresses or IPv4Addresses causes an issue during load. Must include IP,
        # subnet mask and gateway (they can not be patched individually).
        # spec_dict = {'Oem': {self.rdmc.app.typepath.defs.oemhp: {}}}
        spec_dict = dict()
        if "IPv4Addresses" in ethernet_data:
            spec_dict["IPv4Addresses"] = copy.deepcopy(ethernet_data["IPv4Addresses"])
        try:
            if "IPv4Addresses" in ethernet_data["Oem"][self.rdmc.app.typepath.defs.oemhp]:
                spec_dict["Oem"][self.rdmc.app.typepath.defs.oemhp]["IPv4Addresses"] = copy.deepcopy(
                    ethernet_data["Oem"][self.rdmc.app.typepath.defs.oemhp]["IPv4StaticAddresses"]
                )
        except (KeyError, NameError, TypeError, AttributeError):
            pass

        # diff and overwrite the original payload
        ethernet_data = diffdict(ethernet_data, curr_sel.dict)
        ethernet_data.update(spec_dict)

        # verify dependencies on those flags which are to be applied are eliminated
        try:
            # delete Domain name and FQDN if UseDomainName for DHCPv4 or DHCPv6
            # is present. can wait to apply at the end
            if dhcpv4conf.get("UseDomainName"):  # or dhcpv6conf['UseDomainName']:
                if "DomainName" in ethernet_data["Oem"][self.rdmc.app.typepath.defs.oemhp]:
                    del ethernet_data["Oem"][self.rdmc.app.typepath.defs.oemhp]["DomainName"]
                if "FQDN" in ethernet_data:
                    del ethernet_data["FQDN"]
        except (KeyError, NameError, TypeError, AttributeError):
            # try again with OEM
            try:
                if oem_dhcpv4conf.get("UseDomainName") or oem_dhcpv6conf.get("UseDomainName"):
                    if (
                        "Oem" in ethernet_data
                        and "DomainName" in ethernet_data["Oem"][self.rdmc.app.typepath.defs.oemhp]
                    ):
                        del ethernet_data["Oem"][self.rdmc.app.typepath.defs.oemhp]["DomainName"]
                    if "FQDN" in ethernet_data:
                        del ethernet_data["FQDN"]
            except (KeyError, NameError) as exp:
                errors.append("Unable to remove property %s.\n" % exp)

        try:
            # delete DHCP4 DNSServers from IPV4 dict if UseDNSServers Enabled
            # can wait to apply at the end
            if dhcpv4conf.get("UseDNSServers"):  # and ethernet_data.get('NameServers'):
                json_traversal_delete_empty(data=ethernet_data, remove_list=["NameServers"])
        except (KeyError, NameError, TypeError, AttributeError):
            pass
        finally:
            try:
                if oem_dhcpv4conf.get("UseDNSServers"):
                    # del_sections('DNSServers', ethernet_data)
                    json_traversal_delete_empty(data=ethernet_data, remove_list=["DNSServers"])
            except (KeyError, NameError) as exp:
                errors.append("Unable to remove property %s.\n" % exp)
        try:
            if dhcpv4conf.get("UseWINSServers"):
                json_traversal_delete_empty(data=ethernet_data, remove_list=["WINServers"])
        except (KeyError, NameError, TypeError, AttributeError):
            pass
        finally:
            try:
                if oem_dhcpv4conf.get("UseWINSServers"):
                    json_traversal_delete_empty(
                        data=ethernet_data,
                        remove_list=["WINServers", "WINSRegistration"],
                    )
            except (KeyError, NameError) as exp:
                errors.append("Unable to remove property %s.\n" % exp)

        try:
            if dhcpv4conf.get("UseStaticRoutes"):
                json_traversal_delete_empty(data=ethernet_data, remove_list=["StaticRoutes"])
        except (KeyError, NameError, TypeError, AttributeError):
            pass
        finally:
            try:
                if oem_dhcpv4conf.get("UseStaticRoutes"):
                    json_traversal_delete_empty(data=ethernet_data, remove_list=["StaticRoutes"])
            except (KeyError, NameError) as exp:
                errors.append("Unable to remove property %s.\n" % exp)

        try:
            # if using DHCPv4, remove static addresses
            if dhcpv4conf.get("DHCPEnabled"):
                json_traversal_delete_empty(
                    data=ethernet_data,
                    remove_list=["IPv4Addresses", "IPv4StaticAddresses"],
                )
        except (KeyError, NameError, TypeError, AttributeError):
            pass
        finally:
            try:
                if oem_dhcpv4conf.get("Enabled"):
                    json_traversal_delete_empty(
                        data=ethernet_data,
                        remove_list=["IPv4Addresses", "IPv4StaticAddresses"],
                    )
            except (KeyError, NameError) as exp:
                errors.append("Unable to remove property %s.\n" % exp)

        try:
            # if not using DHCPv6, remove static addresses from payload
            if dhcpv6conf.get("OperatingMode") == "Disabled":
                json_traversal_delete_empty(
                    data=ethernet_data,
                    remove_list=["IPv6Addresses", "IPv6StaticAddresses"],
                )
        except (KeyError, NameError, TypeError, AttributeError):
            pass
        finally:
            try:
                if not oem_dhcpv6conf.get("StatefulModeEnabled"):
                    json_traversal_delete_empty(
                        data=ethernet_data,
                        remove_list=["IPv6Addresses", "IPv6StaticAddresses"],
                    )
            except (KeyError, NameError) as exp:
                errors.append("Unable to remove property %s.\n" % exp)

        flags = ethernet_data
        if "StaticNameServers" in flags:
            if dhcpv4curr["UseDNSServers"]:
                flags["DHCPv4"] = {"UseDNSServers": False}

        # verify dependencies on those flags which are to be applied are eliminated
        if (
            "IPv4Addresses" in flags
            and "Address" in flags["IPv4Addresses"][0]
            and "Address" in curr_sel.dict["IPv4Addresses"][0]
            and flags["IPv4Addresses"][0]["Address"] == curr_sel.dict["IPv4Addresses"][0]["Address"]
            and "Gateway" in flags["IPv4Addresses"][0]
            and "Gateway" in curr_sel.dict["IPv4Addresses"][0]
            and flags["IPv4Addresses"][0]["Gateway"] == curr_sel.dict["IPv4Addresses"][0]["Gateway"]
            and "SubnetMask" in flags["IPv4Addresses"][0]
            and "SubnetMask" in curr_sel.dict["IPv4Addresses"][0]
            and flags["IPv4Addresses"][0]["SubnetMask"] == curr_sel.dict["IPv4Addresses"][0]["SubnetMask"]
        ):
            del flags["IPv4Addresses"]
        try:
            if not flags:
                self.rdmc.ui.warn("No change in configurations in " + _path)
            else:
                self.rdmc.app.patch_handler(_path, flags, silent=True)
        except IloResponseError as excp:
            errors.append("iLO Responded with the following errors setting DHCP: %s.\n" % excp)

        try:
            if "AutoNeg" not in list(ethernet_data.keys()):
                json_traversal_delete_empty(data=ethernet_data, remove_list=["FullDuplex", "SpeedMbps"])

            # if Full Duplex exists, check if FullDuplexing enabled. If so,
            # remove Speed setting.
            elif "FullDuplex" in list(ethernet_data.keys()):
                json_traversal_delete_empty(data=ethernet_data, remove_list=["FullDuplex", "SpeedMbps"])
        except (KeyError, NameError) as exp:
            errors.append("Unable to remove property %s.\n" % exp)

        try:
            if "FrameSize" in list(ethernet_data.keys()):
                json_traversal_delete_empty(data=ethernet_data, remove_list=["FrameSize"])
        except (KeyError, NameError) as exp:
            errors.append("Unable to remove property %s.\n" % exp)

        if ethernet_data:
            self.patch_eth(_path, ethernet_data, errors)

        if errors and "Virtual" not in ident_name:
            raise RdmcError(
                "Ethernet configuration errors were found collectively on adapter: "
                "'%s, %s'\ntype: %s\nerrors: %s" % (ident_name, ident_id, _type, errors)
            )

    def patch_eth(self, _path, eth_data, errors=[]):
        """helper function for patching ethernet properties. Retry functionality with the ability
            to remove the offending property.
        :param _path: URI path to be patched.
        :type _path: string
        :param eth_data: JSON containing the ethernet instance (and valid, associated
                         properties) to be patched.
        :type eth_data: JSON
        :param errors: list of errors catalogued between attempts
        :type errors: list

        """

        try:
            if eth_data:
                # eth_data = json.dumps(eth_data)
                # import ast
                # eth_data = ast.literal_eval(eth_data)
                tmp = self.rdmc.app.patch_handler(_path, eth_data, silent=False, service=False)
                if tmp.status == 400:
                    raise InvalidPropertyError(tmp.dict["error"][next(iter(tmp.dict["error"]))])
            else:
                raise NoDifferencesFoundError(
                    "No differences between existing iLO ethernet "
                    "configuration and new ethernet configuration.\nPath: %s\n" % _path
                )

        except InvalidPropertyError as excp:
            errors.append("iLO Responded with the following error: %s.\n" % excp)

            def drill_to_data(data, list_o_keys):
                if len(list_o_keys) > 1:
                    k = list_o_keys.pop(0)
                else:
                    del data[k]
                if isinstance(data, dict):
                    drill_to_data(data[k], list_o_keys)

            if hasattr(excp, "message"):
                for key in excp.message[0]["MessageArgs"]:
                    try:
                        eth_data.pop(key)
                    except (AttributeError, KeyError, StopIteration):
                        try:
                            drill_to_data(
                                eth_data,
                                list_o_keys=json_traversal(eth_data, key, ret_key_path=True),
                            )
                        except:
                            errors.append("Unable to find '%s'" % key)
                        return
                self.patch_eth(_path, eth_data)

        except NoDifferencesFoundError as excp:
            errors.append("%s" % excp)

    def ethernetvalidation(self, options):
        """ethernet validation function
        :param options: command line options
        :type options: list.
        """

        self.cmdbase.login_select_validation(self, options)

        if options.ethfilename:
            if len(options.ethfilename) < 2:
                self.eth_file = options.ethfilename[0]
            else:
                raise InvalidCommandLineError("Only a single ethernet file may be specified.")
        else:
            self.eth_file = __eth_file__

    @staticmethod
    def options_argument_group(parser):
        """Define option arguments group
        :param parser: The parser to add the login option group to
        :type parser: ArgumentParser/OptionParser
        """

        parser.add_argument(
            "--encryption",
            dest="encryption",
            help="Optionally include this flag to encrypt/decrypt a file" " using the key provided.",
            default=None,
        )
        parser.add_argument(
            "-f",
            "--ethfile",
            dest="ethfilename",
            help="""Optionally specify a JSON file to store or load ethernet configuration data.""",
            action="append",
            default=None,
        )

    def definearguments(self, customparser):
        """Wrapper function for certificates command main function
        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)
        self.options_argument_group(customparser)
        subcommand_parser = customparser.add_subparsers(dest="command")
        default_parser = subcommand_parser.add_parser(
            "default",
            help="Obtain iLO management networking interface details and configure basic "
            "properties such as enablement/disablement, domain name servers, ipv4 and ipv6 "
            "networking configuration.",
            formatter_class=RawDescriptionHelpFormatter,
        )
        default_parser.add_argument(
            "--enable_nic",
            dest="enable_nic",
            help="Enable network interfaces. List each interface to be enabled by ID: "
            "Ex: ethernet --enable_nic 1,2,3. **Note**: Non-existent interfaces will be omitted from "
            "configuration.",
            type=str,
            default=None,
        )
        default_parser.add_argument(
            "--disable_nic",
            dest="disable_nic",
            help="Disable network interfaces. List each interface to be disabled by ID: "
            "Ex: ethernet --disable_nic 1,2,3. **Note**: Non-existent interfaces will be omitted from "
            "configuration.",
            type=str,
            default=None,
        )
        default_parser.add_argument(
            "--enable_vnic",
            dest="enable_vnic",
            help="""Enable virtual network interfaces of management network.
                Ex: ethernet --enable_vnic""",
            action="store_true",
            default=False,
        )
        default_parser.add_argument(
            "--disable_vnic",
            dest="disable_vnic",
            help="""Disable virtual network interfaces of management network.
                Ex: ethernet --disable_vnic""",
            action="store_true",
            default=False,
        )
        default_parser.add_argument(
            "--disable_enhanced_downloads",
            dest="disable_enhanced_downloads",
            help="""Disable enhanced download for virtual media and firmware update.
                        Ex: ethernet --disable_enhanced_downloads""",
            action="store_true",
            default=False,
        )
        default_parser.add_argument(
            "--enable_enhanced_downloads",
            dest="enable_enhanced_downloads",
            help="""Enable enhanced download for virtual media and firmware update.
                                Ex: ethernet --enable_enhanced_downloads""",
            action="store_true",
            default=False,
        )
        default_parser.add_argument(
            "--nameservers",
            dest="nameservers",
            help="Configure physical and shared management network interface domain name "
            "servers (DNS) in a list as follows: <DNS1> <DNS2> <DNS3>"
            "Ex: ethernet --nameservers 8.8.8.8,1.1.1.1,2.2.2.2 ethernet --nameservers dns_resolver1.aws.com, "
            "dns_resolver2.aws.com",
            default=None,
        )
        default_parser.add_argument(
            "--proxy",
            dest="proxy",
            type=str,
            help="Configure or clear proxy server for the network "
            "Ex:\nTo set proxy\nethernet --proxy http://proxy.abc.net:8080 or \n"
            "To clear proxy settings\n"
            "ethernet --proxy None",
            default=None,
        )
        default_parser.add_argument(
            "--network_ipv4",
            dest="network_ipv4",
            help="Configure physical and shared management network interface static IPv4 "
            "settings. Settings provided in a list as follows: "
            "<ipv4 address>, <ipv4 gateway>, <ipv4 network mask>. Ex: ethernet --network_ipv4 "
            "192.168.1.10, 192.168.1.1, 255.255.0.0",
            default=None,
        )
        default_parser.add_argument(
            "--network_ipv6",
            dest="network_ipv6",
            help="Configure physical and shared management network interface static IPv6 "
            "settings. Settings provided in a list as follows: "
            "<ipv6 address>, <ipv6 gateway>, <ipv6 network mask>. Ex: ethernet --network_ipv6 "
            "0:0:0:0:ffff:c0a8:10e, 0:0:0:0:0:ffff:c0a8:101, 64. **Note**: IPv6 network mask"
            "is restricted to '64' bits.",
            default=None,
        )
        default_parser.add_argument(
            "-j",
            "--json",
            dest="json",
            action="store_true",
            help="Optionally include this flag if you wish to change the " "displayed output to JSON format.",
            default=False,
        )
        self.cmdbase.add_login_arguments_group(default_parser)
        save_help = "Save a Network Configuration."
        save_parser = subcommand_parser.add_parser(
            __subparsers__[0],
            help=save_help,
            description="{0}\n\texample: ethernet save\n\n\tSave iLO ethernet network management interface "
            "settings to a non-default file name.\n\tEx: ethernet save -f networking.json\n\n\tSave "
            "an encrypted iLO networking configuration file\n\texample: ethernet save --encryption "
            "<ENCRYPTION_KEY>".format(save_help),
            formatter_class=RawDescriptionHelpFormatter,
        )
        self.cmdbase.add_login_arguments_group(save_parser)
        self.options_argument_group(save_parser)
        load_help = "Load a Network Configuration."
        load_parser = subcommand_parser.add_parser(
            __subparsers__[1],
            help=load_help,
            description="{0}\n\texample: ethernet load\n\n\tLoad iLO ethernet networking management interface "
            "settings from a non-default file name.\n\tEx: ethernet load -f "
            "networking.json\n\n\tLoad an encrypted iLO networking configuration file\n\texample: "
            "ethernet load --encryption <ENCRYPTION KEY>".format(load_help),
            formatter_class=RawDescriptionHelpFormatter,
        )
        load_parser.add_argument(
            "--force_network_config",
            dest="force_network_config",
            help="Use this flag to force set network configuration."
            "Network settings will be skipped if the flag is not included.",
            action="store_true",
            default=None,
        )
        self.cmdbase.add_login_arguments_group(load_parser)
        self.options_argument_group(load_parser)



###
# Copyright 2016-2023 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
""" Get Command for RDMC """
from collections import OrderedDict

import six

import redfish.ris
from redfish.ris.utils import iterateandclear

try:
    from rdmc_helper import (
        UI,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        NoContentsFoundForOperationError,
        ReturnCodes,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        UI,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        NoContentsFoundForOperationError,
        ReturnCodes,
    )
try:
    from rdmc_helper import HARDCODEDLIST
except:
    from ilorest.rdmc_helper import HARDCODEDLIST


class GetCommand:
    """Constructor"""

    def __init__(self):
        self.ident = {
            "name": "get",
            "usage": None,
            "description": "To retrieve all"
            " the properties run without arguments. \n\t*Note*: "
            "a type will need to be selected or this will return an "
            "error.\n\texample: get\n\n\tTo retrieve multiple "
            "properties use the following example\n\texample: "
            "get Temperatures/ReadingCelsius Fans/Name --selector=Thermal."
            "\n\n\tTo change output style format provide"
            " the json flag\n\texample: get --json",
            "summary": "Displays the current value(s) of a" " property(ies) within a selected type.",
            "aliases": [],
            "auxcommands": ["LogoutCommand"],
        }
        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()

    def run(self, line, help_disp=False):
        """Main get worker function

        :param line: command line input
        :type line: string.
        """
        if help_disp:
            self.parser.print_help()
            return ReturnCodes.SUCCESS
        try:
            (options, args) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")

        if getattr(options, "json"):
            self.rdmc.json = True

        selector = options.selector.lower() if options.selector else self.rdmc.app.selector.lower()

        if "securityservice" in selector:
            self.cmdbase.login_validation(self, options)
        else:
            self.getvalidation(options)

        filtr = (None, None)
        if options.filter:
            try:
                if (str(options.filter)[0] == str(options.filter)[-1]) and str(options.filter).startswith(("'", '"')):
                    options.filter = options.filter[1:-1]

                (sel, val) = options.filter.split("=")
                filtr = (sel.strip(), val.strip())

            except:
                raise InvalidCommandLineError("Invalid filter" " parameter format [filter_attribute]=[filter_value]")

        self.getworkerfunction(
            args,
            options,
            results=None,
            uselist=True,
            filtervals=filtr,
            readonly=options.noreadonly,
        )

        self.cmdbase.logout_routine(self, options)
        # Return code
        return ReturnCodes.SUCCESS

    def getworkerfunction(
        self,
        args,
        options,
        readonly=False,
        filtervals=(None, None),
        results=None,
        uselist=False,
    ):
        """main get worker function

        :param args: command line arguments
        :type args: list.
        :param options: command line options
        :type options: list.
        :param line: command line input
        :type line: string.
        :param readonly: remove readonly properties
        :type readonly: bool
        :param filtervals: filter key value pair (Key,Val)
        :type filtervals: tuple
        :param results: current results collected
        :type results: string.
        :param uselist: use reserved properties list to filter results
        :type uselist: boolean.
        """
        nocontent = set()
        instances = None
        arg = None

        # For rest redfish compatibility of bios.
        if hasattr(options, "selector") and options.selector:
            self.rdmc.app.selector = options.selector
        args = [args] if args and isinstance(args, six.string_types) else args
        if self.rdmc.app.selector and "." not in self.rdmc.app.selector:
            self.rdmc.app.selector = self.rdmc.app.selector + "."
        args = (
            [
                (
                    "Attributes/" + arg
                    if self.rdmc.app.selector
                    and self.rdmc.app.selector.lower().startswith("bios.")
                    and "attributes" not in arg.lower()
                    else arg
                )
                for arg in args
            ]
            if args
            else args
        )
        if filtervals[0]:
            instances = self.rdmc.app.select(selector=self.rdmc.app.selector, fltrvals=filtervals)

        try:
            if "securityservice" in self.rdmc.app.selector.lower():
                url = "/redfish/v1/Managers/1/SecurityService/"
                contents = self.rdmc.app.get_handler(url, service=True, silent=True).dict
                security_contents = []
                if not args and not options.json:
                    UI().print_out_human_readable(contents)
                elif options and options.json and contents:
                    UI().print_out_json(contents)
                else:
                    attr = args[0]
                    contents_lower = {k.lower(): v for k, v in contents.items()}
                    security_contents.append({attr: contents_lower[attr.lower()]})
                if security_contents:
                    UI().print_out_human_readable(security_contents)
            else:
                if "selector" in options:
                    contents = self.rdmc.app.getprops(
                        props=args,
                        remread=readonly,
                        selector=options.selector,
                        nocontent=nocontent,
                        insts=instances,
                    )
                else:
                    contents = self.rdmc.app.getprops(
                        props=args, remread=readonly, nocontent=nocontent, insts=instances
                    )
            uselist = False if readonly else uselist
        except redfish.ris.rmc_helper.EmptyRaiseForEAFP:
            contents = self.rdmc.app.getprops(props=args, nocontent=nocontent)
        for ind, content in enumerate(contents):
            if "bios." in self.rdmc.app.selector.lower() and "Attributes" in list(content.keys()):
                content.update(content["Attributes"])
                del content["Attributes"]
            contents[ind] = OrderedDict(sorted(list(content.items()), key=lambda x: x[0]))
        if len(args) > 0 and "Members" in args[0]:
            uselist = False
        if uselist:
            for item in contents:
                self.removereserved(item)
        if results:
            return contents

        contents = contents[0] if len(contents) == 1 else contents

        if options and options.json and contents:
            UI().print_out_json(contents)
        elif contents:
            UI().print_out_human_readable(contents)
        else:
            try:
                if nocontent or not any(next(iter(contents))):
                    raise Exception()
            except Exception:
                strtoprint = ", ".join(str(val) for val in nocontent)
                if not strtoprint and arg:
                    strtoprint = arg
                    raise NoContentsFoundForOperationError("No get contents found for entry: %s" % strtoprint)
                else:
                    raise NoContentsFoundForOperationError("No get contents found for " "selected type.")
        if options.logout:
            self.auxcommands["logout"].run("")

    def removereserved(self, entry):
        """function to remove reserved properties
        :param entry: dictionary to remove reserved properties from
        :type entry: dict.
        """

        for key, val in list(entry.items()):
            if key.lower() in HARDCODEDLIST or "@odata" in key.lower():
                del entry[key]
            elif isinstance(val, list):
                for item in entry[key]:
                    if isinstance(item, dict) and item:
                        self.removereserved(item)
                        if all([True if not test else False for test in entry[key]]):
                            del entry[key]
            elif isinstance(val, dict):
                self.removereserved(val)
                if all([True if not test else False for test in entry[key]]):
                    del entry[key]

    def checktoprint(self, options, contents, nocontent, arg):
        """function to decide what/how to print
        :param options: list of options
        :type options: list.
        :param contents: dictionary value returned by getprops.
        :type contents: dict.
        :param nocontent: props not found are added to the list.
        :type nocontent: list.
        :param arg: string of args
        :type arg: string
        """
        if options and options.json and contents:
            self.rdmc.ui.print_out_json(contents)
        elif contents:
            self.rdmc.ui.print_out_human_readable(contents)
        else:
            try:
                if nocontent or not any(next(iter(contents))):
                    raise Exception()
            except:
                strtoprint = ", ".join(str(val) for val in nocontent)
                if not strtoprint and arg:
                    strtoprint = arg
                    raise NoContentsFoundForOperationError("No get contents " "found for entry: %s" % strtoprint)
                else:
                    raise NoContentsFoundForOperationError("No get contents " "found for selected type.")

    def collectandclear(self, contents, key, values):
        """function to find and remove unneeded values from contents dictionary
        :param contents: dictionary value returned by getprops
        :type contents: dict.
        :param key: string of keys
        :type key: string.
        :param values: list of values
        :type values: list.
        """
        clearcontent = contents[0][key]
        if isinstance(clearcontent, dict):
            keyslist = list(clearcontent.keys())
        else:
            keyslist = [clearcontent]
        clearedlist = keyslist
        for arg in values:
            for keys in keyslist:
                if str(keys).lower() == str(arg).lower():
                    clearedlist.remove(arg)
        contents = iterateandclear(contents, clearedlist)
        return contents

    def getvalidation(self, options):
        """get method validation function

        :param options: command line options
        :type options: list.
        """
        self.cmdbase.login_select_validation(self, options)

    def definearguments(self, customparser):
        """Wrapper function for new command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)

        customparser.add_argument(
            "--selector",
            dest="selector",
            help="Optionally include this flag to select a type to run"
            " the current command on. Use this flag when you wish to"
            " select a type without entering another command, or if you"
            " wish to work with a type that is different from the one"
            " you currently have selected.",
            default=None,
        )

        customparser.add_argument(
            "--filter",
            dest="filter",
            help="Optionally set a filter value for a filter attribute."
            " This uses the provided filter for the currently selected"
            " type. Note: Use this flag to narrow down your results. For"
            " example, selecting a common type might return multiple"
            " objects that are all of that type. If you want to modify"
            " the properties of only one of those objects, use the filter"
            " flag to narrow down results based on properties."
            "\t\t\t\t\t Usage: --filter [ATTRIBUTE]=[VALUE]",
            default=None,
        )
        customparser.add_argument(
            "-j",
            "--json",
            dest="json",
            action="store_true",
            help="Optionally include this flag if you wish to change the"
            " displayed output to JSON format. Preserving the JSON data"
            " structure makes the information easier to parse.",
            default=False,
        )
        customparser.add_argument(
            "--noreadonly",
            dest="noreadonly",
            action="store_true",
            help="Optionally include this flag if you wish to only show"
            " properties that are not read-only. This is useful to see what "
            "is configurable with the selected type(s).",
            default=False,
        )
        customparser.add_argument(
            "--refresh",
            dest="ref",
            action="store_true",
            help="Optionally reload the data of selected type and clear " "patches from current selection.",
            default=False,
        )
